#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides pooling functions for Animator.
"""
import socket
import numpy as np
from typing import Tuple, List, Any, NamedTuple, Dict, Optional
from scine_heron.mediator_potential import clientserver


class GradientCalculationResult(NamedTuple):
    """
    The results of a Sparrow calculation.
    """

    gradients: np.ndarray
    energy: float
    atomic_charges: Optional[List[float]]
    settings: Dict[str, Any]
    bond_orders: Optional[np.ndarray]
    error_msg: str
    info_msg: str
    molden_input: str


def calculate_gradient(
    parameters: Tuple[
        int, List[Tuple[str, Tuple[float, float, float]]], Tuple[str, str], Dict[str, Any], bool, str
    ]
) -> GradientCalculationResult:
    molecule_version, molecule, calculator_args, settings, mediator_potential_signal, bond_type = parameters
    stop_signal = False
    new_settings = settings
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.connect(("127.0.0.1", 55145))
            clientserver.send_data(
                data=[
                    stop_signal,
                    molecule_version,
                    molecule,
                    calculator_args,
                    settings,
                    mediator_potential_signal,
                    bond_type,
                ],
                socket=s,
            )
            (
                energy,
                gradients,
                charges,
                bond_orders,
                molden_input,
                error_msg,
                info_msg,
                new_settings,
                mediator_potential_signal
            ) = clientserver.recv_data(connection=s)
        except ConnectionError:
            energy, gradients, charges, bond_orders, molden_input, error_msg, info_msg = (
                0.0,
                np.zeros(shape=(len(molecule), 3)),
                None,
                None,
                None,
                "",
                "",
            )
        if new_settings != settings:
            settings = new_settings
        if bond_orders is not None:
            # nd.array is not JSON serializable, so we convert it back from a list
            bond_orders = np.array(bond_orders)
    return GradientCalculationResult(
        gradients=np.array(gradients),
        energy=energy,
        settings=settings,
        atomic_charges=charges,
        bond_orders=bond_orders,
        error_msg=error_msg,
        info_msg=info_msg,
        molden_input=molden_input,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Callable, Optional

from PySide2.QtWidgets import QPushButton, QWidget
from PySide2.QtGui import QKeySequence

from .clickable import Clickable


class TextPushButton(QPushButton, Clickable):
    """
    A push button that triggers a function and only displays some text.
    """

    def __init__(self,
                 text: str,
                 action: Optional[Callable[[], None]] = None,
                 parent: Optional[QWidget] = None,
                 shortcut: Optional[str] = None,
                 max_width: Optional[int] = None) -> None:
        """
        Construct the button with the displayed text and the function to be called
        when the button is clicked

        Parameters
        ----------
        text : str
            The displayed text
        action : Callable
            The function triggerd by clicking
        parent : Optional[QWidget], optional
            The parent widget, by default None
        shortcut : Optional[str], optional
            A keyboard shortcut for the button, which is also added as a tooltip, by default None
        """
        super().__init__(text, parent=parent)
        if action is not None:
            self.clicked.connect(action)  # pylint: disable=no-member
        if shortcut is not None:
            self.setShortcut(QKeySequence(shortcut))
            self.setToolTip(shortcut)
        if max_width is not None:
            self.setMaximumWidth(max_width)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the SettingsWidget class.
"""
from pkgutil import iter_modules
from typing import List, Dict, Any, Callable
from functools import partial

from PySide2.QtCore import QSize, Qt
from PySide2.QtWidgets import (
    QDockWidget,
    QWidget,
    QSpinBox,
    QLabel,
    QDoubleSpinBox,
    QCheckBox,
)

from scine_heron.containers.combo_box import BaseBox
from scine_heron.containers.layouts import HorizontalLayout, VerticalLayout
from scine_heron.containers.buttons import TextPushButton
from scine_heron.settings.settings import MoleculeStyle, LabelsStyle
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.status_manager import (
    Status,
    StatusManager,
)


class BasicSettingsWidget(QDockWidget):
    """
    Displays of display and molecular orbital settings
    """

    def __init__(self, settings_status_manager: SettingsStatusManager):
        """
        This is the display of the settings grid. Each setting is drawn depending on its type.
        :param settings_status_manager: class managing MO and display settings and propagating them
        """
        QDockWidget.__init__(self, "Basic Settings")
        self.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.setMinimumWidth(500)

        self.__dockedWidget = QWidget(self)
        self.setWidget(self.__dockedWidget)

        self.__layout = VerticalLayout()
        self.__dockedWidget.setLayout(self.__layout)
        self.__layout.setAlignment(Qt.AlignTop)

        self.__settings_status_manager = settings_status_manager
        self.__settings_status_manager.number_of_mos_changed.connect(
            self.__update_number_of_mo
        )
        self.__settings_status_manager.selected_mo_changed.connect(
            self.__update_number_of_mo
        )

        self.__widgets_dict: Dict[str, Any] = {}

        self.__widget_height = 30
        self.__widget_width = 160

        self.__enabled = StatusManager(True)

        self.__add_style_setting(
            "Molecule Style",
            "molecule_style",
            [str(m.value) for m in MoleculeStyle.__members__.values()],
            self.__update_molecule_style,
            str(MoleculeStyle.BallAndStick.value),
            enabled=self.__enabled,
        )

        self.__add_style_setting(
            "Labels Style",
            "labels_style",
            [str(m.value) for m in LabelsStyle.__members__.values()],
            self.__update_labels_style,
            str(LabelsStyle.Empty.value),
            enabled=self.__enabled,
        )
        self.__add_style_setting(
            "Bond Display",
            "bond_display",
            ["distance", "el. density"],
            self.__update_bond_style,
            "distance",
            enabled=self.__enabled,
        )

        self.__add_mo_setting_at_layout(
            "Molecular Orbital", "molecular_orbital", self.__enabled
        )
        self.__add_double_spin_setting_at_layout(
            "Contour Value",
            "molecular_orbital_value",
            0.05,
            self.__update_molecular_orbital_value,
            self.__enabled,
        )
        self.__add_double_spin_setting_at_layout(
            "Scale Haptic Force",
            "scale_force",
            1.0,
            self.__update_force_scaling,
            self.__enabled,
            0.6,
            1.4,
        )
        self.__add_double_spin_setting_at_layout(
            "Scale applied gradients",
            "scale_gradients",
            0.1,
            self.__update_gradient_scaling,
            self.__enabled,
            0.01,
            10.0,
        )
        mediator_check = QCheckBox()
        mediator_check.setChecked(True)
        self.__enabled.changed_signal.connect(mediator_check.setEnabled)
        mediator_check.stateChanged.connect(self.__update_mediator_state)  # pylint: disable=no-member
        self.__widgets_dict["mediator_potential_active"] = mediator_check
        self.__layout.addLayout(HorizontalLayout([QLabel("Active mediator potential"), mediator_check]))

        if any(name == "scine_swoose" for _, name, __ in iter_modules()):
            from scine_heron.settings.swoose_settings import SwooseSettingsWidget
            swoose_settings_widget = SwooseSettingsWidget(self, self.__settings_status_manager)
            self.__layout.addWidget(swoose_settings_widget)

    def __update_number_of_mo(self) -> None:
        widget = self.__widgets_dict["molecular_orbital"]
        widget.setMinimum(1)

        if self.__settings_status_manager.number_of_molecular_orbital is not None:
            widget.setMaximum(
                self.__settings_status_manager.number_of_molecular_orbital
            )
        selected = self.__settings_status_manager.selected_molecular_orbital
        if selected is not None and selected > 0:
            widget.setValue(selected)

    def set_enabled(self, enabled: bool) -> None:
        """
        Set widgets enabled.
        """
        self.__enabled.value = enabled

    def is_enabled(self) -> bool:
        """
        Return whether widgets are enabled.
        """
        return self.__enabled.value

    def __add_mo_setting_at_layout(
        self, setting_name: str, setting_key: str, enabled: Status[bool],
    ) -> None:
        """
        Add QSpinBox widget for molecular orbital setting.
        setting_name is a setting display name.
        setting_key is a setting name in sparrow.
        """
        spin_edit = QSpinBox()
        spin_edit.setFixedSize(spin_edit.sizeHint())
        spin_edit.setMinimum(1)
        spin_edit.setValue(1)
        spin_edit.setMaximum(1)
        self.__widgets_dict[setting_key] = spin_edit
        enabled.changed_signal.connect(spin_edit.setEnabled)

        homo_button = TextPushButton("HOMO", partial(self.__update_molecular_orbital, setting_key, -1), self)
        enabled.changed_signal.connect(homo_button.setEnabled)

        lumo_button = TextPushButton("LUMO", partial(self.__update_molecular_orbital, setting_key, -2), self)
        enabled.changed_signal.connect(lumo_button.setEnabled)

        density_button = TextPushButton("El. Density", partial(self.__update_molecular_orbital, setting_key, -3), self)
        enabled.changed_signal.connect(density_button.setEnabled)

        calculate_button = TextPushButton("Go", partial(self.__update_molecular_orbital, setting_key), self)
        calculate_button.setFixedSize(calculate_button.sizeHint())
        enabled.changed_signal.connect(calculate_button.setEnabled)

        self.__layout.add_layouts([
            HorizontalLayout([QLabel(setting_name)]),
            HorizontalLayout([homo_button, lumo_button]),
            HorizontalLayout([density_button]),
            HorizontalLayout([QLabel("MO Index"), spin_edit, calculate_button]),
        ])

    def __add_double_spin_setting_at_layout(
        self,
        setting_name: str,
        setting_key: str,
        default_value: float,
        update_function: Callable[[float], None],
        enabled: Status[bool],
        min_value: float = -1000000000.0,
        max_value: float = 1000000000.0,
    ) -> QDoubleSpinBox:
        """
        Add QDoubleSpinBox widget.
        setting_name is a setting display name.
        setting_key is a setting name in sparrow.
        default_value is a default display value.
        """
        spin_edit = QDoubleSpinBox()
        spin_edit.setFixedSize(QSize(self.__widget_width, self.__widget_height))
        spin_edit.setMinimum(min_value)
        spin_edit.setMaximum(max_value)
        spin_edit.setValue(default_value)
        spin_edit.setSingleStep(0.01)
        spin_edit.setDecimals(2)
        enabled.changed_signal.connect(spin_edit.setEnabled)

        spin_edit.valueChanged.connect(update_function)  # pylint: disable=no-member

        self.__widgets_dict[setting_key] = spin_edit

        self.__layout.addLayout(HorizontalLayout([QLabel(setting_name), spin_edit]))

        return spin_edit

    def __add_style_setting(
        self,
        setting_name: str,
        setting_key: str,
        all_values: List[str],
        update: Callable[[List[str], int], None],
        default_value: str,
        enabled: Status[bool],
    ) -> None:
        """
        Add molecule style widget.
        setting_name is a setting display name.
        setting_key is a setting name in sparrow.
        all_values is a list of valid values.
        default_value is a default display value.
        """
        combo_box = self.__add_combo_box_at_layout(
            setting_name, setting_key, all_values
        )
        combo_box.currentIndexChanged.connect(partial(update, all_values))  # pylint: disable=no-member
        combo_box.setCurrentIndex(all_values.index(default_value))
        enabled.changed_signal.connect(combo_box.setEnabled)

    def __add_combo_box_at_layout(
        self, setting_name: str, setting_key: str, all_values: List[str],
    ) -> BaseBox:
        """
        Add QComboBox widget.
        setting_name is a setting display name.
        setting_key is a setting name in sparrow.
        all_values is a list of valid values.
        """
        combo_box = BaseBox()
        combo_box.addItems(all_values)
        combo_box.setFixedSize(QSize(self.__widget_width, self.__widget_height + 1))

        self.__widgets_dict[setting_key] = combo_box

        self.__layout.addLayout(HorizontalLayout([QLabel(setting_name), combo_box]))

        return combo_box

    def __update_molecule_style(self, all_values: List[str], index: int) -> None:
        """
        Update molecule style.
        """
        self.__settings_status_manager.molecule_style = MoleculeStyle(all_values[index])

    def __update_molecular_orbital(
        self, setting_key: str, special_case: int = 0
    ) -> None:
        """
        Update molecule orbital.
        """
        if special_case:
            self.__settings_status_manager.selected_molecular_orbital = special_case
        else:
            self.__settings_status_manager.selected_molecular_orbital = self.__widgets_dict[
                setting_key
            ].value()

    def __update_molecular_orbital_value(self, value: float) -> None:
        """
        Update molecule orbital value.
        """
        self.__settings_status_manager.molecular_orbital_value = value

    def __update_force_scaling(self, value: float) -> None:
        """
        Update scaling for haptic feedback.
        """
        self.__settings_status_manager.haptic_force_scaling = value

    def __update_gradient_scaling(self, value: float) -> None:
        """
        Update scaling of applied gradients.
        """
        self.__settings_status_manager.gradients_scaling = value

    def __update_labels_style(self, all_values: List[str], index: int) -> None:
        """
        Update molecule style.
        """
        self.__settings_status_manager.labels_style = LabelsStyle(all_values[index])

    def __update_bond_style(self, all_values: List[str], index: int) -> None:
        """
        Update criterion for drawing bonds.
        """
        self.__settings_status_manager.bond_display = all_values[index]

    def __update_mediator_state(self, value: bool) -> None:
        """
        Update if the mediator potential should be active.

        Parameters
        ----------
        value : bool
            The new value.
        """
        self.__settings_status_manager.set_mediator_potential_setting(value)

    def get_mediator_active(self) -> bool:
        return self.__settings_status_manager.get_mediator_potential_setting()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the class Animator.
"""

from typing import Any, Optional, Callable, TypeVar, TYPE_CHECKING
from concurrent.futures import ProcessPoolExecutor, Executor
from PySide2.QtCore import QObject, QTimer
from .play_status_manager import PlayStatusManager
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class Animator(QObject):
    """
    Animates an entity by updating it in intervals.
    """

    render_signal = Signal()

    S = TypeVar("S")
    T = TypeVar("T")

    def __init__(
        self,
        update: Callable[[S], T],
        provide_args: Callable[[], S],
        apply_results: Callable[[T, float], None],
        pool: Executor = ProcessPoolExecutor(1),
        parent: Optional[QObject] = None,
        interval: int = 1000 // 60,
    ):
        """
        Create an animator by providing an `update` function that is
        intended to be called every `interval` ms. The calls
        to `update` are submitted to `pool`.
        """
        super().__init__(parent)

        self.__update = update
        self.__provide_args = provide_args
        self.__apply_results = apply_results

        self.__pool = pool

        self.__timer = QTimer()
        self.__timer.setInterval(interval)
        self.__timer.timeout.connect(self.__tick)  # pylint: disable=no-member

        self.__status_manager = PlayStatusManager()
        self.__status_manager.on_signal.connect(self.__timer.start)
        self.__status_manager.off_signal.connect(self.__timer.stop)

    def __tick(self) -> None:
        """
        Blocks further signals from the timer and submits an update.
        """
        self.__timer.blockSignals(True)

        try:
            future = self.__pool.submit(self.__update, self.__provide_args())
            future.add_done_callback(self.__callback)
        except Exception as exception:
            self.__timer.blockSignals(False)
            raise exception

    def __callback(self, calculation_result) -> None:
        """
        Called when an update finished. Unblocks the timer signals
        and emits a `render_signal`.
        """
        try:
            self.__apply_results(
                calculation_result.result(), self.__timer.interval() / 1000,
            )
        finally:
            self.render_signal.emit()
            self.__timer.blockSignals(False)

    def stop(self) -> None:
        """
        Stops the animation.
        """
        self.__status_manager.stop()

    def start(self) -> None:
        """
        Starts the animation.
        """
        self.__status_manager.start()

    @property
    def running(self) -> bool:
        return self.__status_manager.is_on()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtWidgets import QTabWidget
from PySide2.QtGui import QCursor
from PySide2.QtCore import Qt


class TabWidget(QTabWidget):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.tabBar().setCursor(QCursor(Qt.CursorShape.PointingHandCursor))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the EnergyProfileStatusManager class.
"""

from typing import Any, List, Optional, TYPE_CHECKING
from datetime import datetime
from PySide2.QtCore import QObject
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class EnergyProfileStatusManager(QObject):
    """
    Provides a Status Manager that hold a list and emits events on resize
    """

    changed_signal = Signal(list)

    def __init__(self, *args: Any):
        super().__init__(*args)
        self.value: List[Any] = []

    def append(self, item: Any) -> None:
        """
        Appends data to inner list
        """
        self.value.append(item)
        self.changed_signal.emit(self.value)

    def __len__(self) -> int:
        """
        Returns the len of the inner list
        """
        return len(self.value)

    def reset(self) -> None:
        """
        Empties the inner list
        """
        self.value = []

    def get_latest_energy(self, seconds) -> Optional[float]:
        if len(self.value) > 0:
            if (datetime.now() - self.value[-1].time_stamp).total_seconds() < seconds:
                return self.value[-1].energy
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from enum import Enum
from typing import Type, Optional

from PySide2.QtCore import QObject
from scine_heron.containers.combo_box import ComboBox


class EnumSelectionBox(ComboBox):
    """
    A combo box that can be constructed with a ChemotonClassSearcher to include all its
    held classes as selectable items and split them based on its module information.
    """

    def __init__(self, parent: Optional[QObject], enum_type: Type[Enum], add_none: bool = False) -> None:
        """
        Construct the widget that allows to select any member of the Enum type

        Parameters
        ----------
        parent : QObject
            The parent widget
        enum_type : Type
            The enum type to select from, class must inherit from Enum
        add_none : bool, optional
            If 'None' should be added as an option to the combo box, by default False
        """
        self.enum_values = list(enum_type.__members__.values())
        if not self.enum_values:
            raise ValueError(f"Enum type {enum_type.__name__} has no members")

        ComboBox.__init__(self, parent=parent, values=list(enum_type.__members__.keys()), header=enum_type.__name__,
                          add_none=add_none)

    def get_value(self) -> Enum:
        """
        Returns the selected enum value
        """
        current_text = self.currentText().lower()
        for value in self.enum_values:
            if value.name.lower() == current_text:
                return value
        raise ValueError(f"Could not find enum value for {current_text}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the System class.
"""

from typing import Any, Dict, List, Tuple
from numpy import ndarray


class System:
    # This class contains the current positions and elements of the system
    # Only altered by the server process, only read by the sparrow process
    # All in atomic units

    def __init__(
        self,
        molecule_version: int,
        positions: ndarray,
        atom_symbols: List[str],
        calculator_args: Tuple[str, str],
        settings: Dict[str, Any],
    ):
        self.__molecule_version = molecule_version
        self.__positions = positions
        self.__atom_symbols = atom_symbols
        self.__calculator_args = calculator_args
        self.__settings = settings

    @property
    def molecule_version(self) -> int:
        return self.__molecule_version

    @property
    def positions(self) -> ndarray:
        return self.__positions

    @property
    def atom_symbols(self) -> List[str]:
        return self.__atom_symbols

    @property
    def settings(self) -> Dict[str, Any]:
        return self.__settings

    @property
    def calculator_args(self) -> Tuple[str, str]:
        return self.__calculator_args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtCore import Qt
from PySide2.QtWidgets import QStyledItemDelegate, QStyle
from PySide2.QtGui import QColor


class CustomLightDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        if option.state & QStyle.State_Selected:
            painter.fillRect(option.rect, QColor("#215CAF"))
        else:
            painter.fillRect(option.rect, QColor("#4D7DBF"))

        painter.setPen(QColor("#ffffff"))
        painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, index.data())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import List, Any, Tuple, Union, Dict, Set, Optional

import scine_database as db
import scine_utilities as utils
from scine_database.concentration_query_functions import query_concentration, query_concentrations

from scine_heron.molecule.molecule_widget import MoleculeWidget

from PySide2.QtWidgets import (
    QWidget,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QHBoxLayout,
    QLineEdit,
    QCheckBox,
    QScrollArea,
)

from PySide2.QtCore import Qt

from scine_heron.containers.combo_box import BaseBox


class DirectedExplorationProgressView(QScrollArea):
    def __init__(self, db_manager: db.Manager) -> None:
        super().__init__()

        self.db_manager: db.Manager = db_manager
        self.__currently_updating = False
        self._old_layout: Union[None, QVBoxLayout] = None

        self._element_width = 200
        self._element_height = 200

    def update_view(self, step_to_compound_mapper: Any):
        if self.__currently_updating:
            return
        self.__currently_updating = True
        compound_collection = self.db_manager.get_collection("compounds")
        structure_collection = self.db_manager.get_collection("structures")
        # Remove old stuff
        if self._old_layout:
            QWidget().setLayout(self._old_layout)
        layout = QVBoxLayout()
        step_to_compound_map = step_to_compound_mapper.get_step_to_compound_mapping(
            self.parent().settings.get_comparison_threshold())
        show_only_added_compounds: bool = self.parent().settings.show_only_added_compounds
        already_present_compounds: Set[str] = set()
        for i, compound_id_list in enumerate(step_to_compound_map):
            step_widget = QWidget()
            qhbox = QHBoxLayout()
            self.add_description_box(qhbox, i + 1)
            added_compounds = False
            for c_id, label in compound_id_list:
                if show_only_added_compounds and i > 0:
                    if c_id.string() in already_present_compounds:
                        continue
                added_compounds = True
                compound = db.Compound(c_id, compound_collection)
                centroid = db.Structure(compound.get_centroid(), structure_collection)
                self.add_molecule(centroid, label, c_id, qhbox)
            if i == 0:
                already_present_compounds = set(c_id.string() for c_id, _ in compound_id_list)
            else:
                already_present_compounds = already_present_compounds.union(
                    set(c_id.string() for c_id, _ in compound_id_list))
            if added_compounds:
                step_widget.setLayout(qhbox)
                scroll_area = QScrollArea()
                scroll_area.setFixedWidth(self.width() - 25)
                scroll_area.setWidget(step_widget)
                layout.addWidget(scroll_area)

        self._old_layout = layout
        view_widget = QWidget()
        view_widget.setLayout(layout)
        self.setWidget(view_widget)

        self.__currently_updating = False

    def add_description_box(self, layout: QHBoxLayout, step_index: int):
        qvbox = QVBoxLayout()
        q_label = QLabel()
        q_label.setText("Step " + str(step_index))
        q_label.setFixedSize(int(self._element_width / 2), self._element_height)
        qvbox.addWidget(q_label, Qt.AlignTop)

        q_label1 = QLabel()
        q_label1.setText("c_max\nc_final\nc_flux\nID")
        q_label1.setFixedSize(int(self._element_width / 2), int(self._element_height / 2))
        qvbox.addWidget(q_label1, Qt.AlignTop)
        layout.addLayout(qvbox, Qt.AlignTop)

    def add_molecule(self, structure: db.Structure, label: str, c_id: db.ID, layout: QHBoxLayout):
        qvbox = QVBoxLayout()
        new_widget = MoleculeWidget(
            self, atoms=structure.get_atoms(),
        )
        new_widget.setFixedSize(self._element_width, self._element_height)
        qvbox.addWidget(new_widget, Qt.AlignTop)

        q_label = QLabel()
        q_label.setText(label + "\n" + c_id.string())
        q_label.setFixedSize(self._element_width, int(self._element_height / 2))
        q_label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        qvbox.addWidget(q_label, Qt.AlignTop)
        layout.addLayout(qvbox, Qt.AlignTop)


class KineticModelingDrivenExplorationExtractor:
    def __init__(self, db_manager: db.Manager):
        self._kinetic_modeling_job_order: str = "kinetx_kinetic_modeling"
        self._aggregate_id_key: str = "aggregate_ids"
        self._aggregate_type_key: str = "aggregate_types"
        self._start_concentrations_key: str = "start_concentrations"
        self._structures: db.Collection = db_manager.get_collection("structures")
        self._compounds: db.Collection = db_manager.get_collection("compounds")
        self._calculations: db.Collection = db_manager.get_collection("calculations")
        self._properties: db.Collection = db_manager.get_collection("properties")
        self.max_concentration_label: str = "max_concentration"
        self.final_concentration_label: str = "final_concentration"
        self.concentration_flux_label: str = "concentration_flux"
        self._comparison_index: int = 0
        self._concentrations_to_compounds_map: Dict[str, Tuple[List[float], List[float], List[float]]] = dict()
        self._n_calcs_last: int = 0
        self._compound_ids_per_calculations: List[List[Tuple[db.ID, str]]] = list()
        self._old_threshold: Union[None, float] = None
        self._concentrations_to_calculations_map: Dict[str, List[int]] = dict()
        self._starting_compound_ids: Optional[List[Tuple[db.ID, str]]] = None

    def reset_cache(self):
        self._n_calcs_last = 0
        self._concentrations_to_compounds_map = dict()
        self._old_threshold = None

    def set_comparison_index(self, new_index: int):
        self._comparison_index = new_index
        self.reset_cache()

    def get_comparison_options(self):
        return [self.max_concentration_label, self.final_concentration_label, self.concentration_flux_label]

    def get_step_to_compound_mapping(self, comparison_threshold: float) -> List[List[Tuple[db.ID, str]]]:
        return self._get_compounds_in_kinetic_modeling_jobs(comparison_threshold)

    def _get_one_input_structure(self):
        # Check for a compound with a start concentration
        for compound in self._compounds.iterate_all_compounds():
            compound.link(self._compounds)
            centroid = db.Structure(compound.get_centroid(), self._structures)
            start_concentration = query_concentration("start_concentration", centroid, self._properties)
            if start_concentration:
                return centroid

    def _get_initial_compounds_from_calculation_settings(self, settings: utils.ValueCollection)\
            -> List[Tuple[db.ID, str]]:
        if not self._starting_compound_ids:
            if settings[self._start_concentrations_key]:
                assert isinstance(settings[self._start_concentrations_key], list)
            if settings[self._aggregate_id_key]:
                assert isinstance(settings[self._aggregate_id_key], list)
            if settings[self._aggregate_type_key]:
                assert isinstance(settings[self._aggregate_type_key], list)
            start_concentrations: List[float] = settings[self._start_concentrations_key]  # type: ignore
            aggregate_ids: List[str] = settings[self._aggregate_id_key]  # type: ignore
            aggregate_types: List[int] = settings[self._aggregate_type_key]  # type: ignore
            self._starting_compound_ids = list()
            for concentration, a_str_id, a_type in zip(start_concentrations, aggregate_ids, aggregate_types):
                if concentration > 0.0 and db.CompoundOrFlask(a_type) == db.CompoundOrFlask.COMPOUND:
                    self._starting_compound_ids.append((db.ID(a_str_id), "\ninput\n"))
        return self._starting_compound_ids

    def _keep_track_on_concentration_to_calculation_mapping(self, str_id: str, calc_index: int):
        if str_id in self._concentrations_to_calculations_map:
            self._concentrations_to_calculations_map[str_id].append(calc_index)
        else:
            self._concentrations_to_calculations_map[str_id] = [calc_index]

    def _get_concentration_index_for_calculation(self, str_id: str, calc_index: int):
        return self._concentrations_to_calculations_map[str_id].index(calc_index)

    def _get_compounds_in_kinetic_modeling_jobs(self, comparison_threshold: float) -> List[List[Tuple[db.ID, str]]]:
        calculation_ids = self._get_one_input_structure().get_calculations(self._kinetic_modeling_job_order)
        if len(calculation_ids) < 1:
            return list()
        if len(calculation_ids) > self._n_calcs_last:
            self.reset_cache()
        if self._old_threshold:
            if abs(self._old_threshold - comparison_threshold) < 1e-14:
                return self._compound_ids_per_calculations

        self._n_calcs_last = len(calculation_ids)
        settings_0 = db.Calculation(calculation_ids[0], self._calculations).get_settings()
        self._compound_ids_per_calculations = [self._get_initial_compounds_from_calculation_settings(settings_0)]
        self._concentrations_to_compounds_map = dict()
        for i, calc_id in enumerate(calculation_ids):
            calculation = db.Calculation(calc_id, self._calculations)
            settings = calculation.get_settings()
            compound_ids: List[Tuple[db.ID, str]] = list()
            if calculation.get_status() != db.Status.COMPLETE:
                continue
            a_str_ids: List[str] = settings[self._aggregate_id_key]  # type: ignore
            a_int_types: List[int] = settings[self._aggregate_type_key]  # type: ignore
            for a_str_id, a_type in zip(a_str_ids, a_int_types):
                if db.CompoundOrFlask(a_type) == db.CompoundOrFlask.COMPOUND:
                    c_id = db.ID(a_str_id)
                    compound = db.Compound(c_id, self._compounds)
                    self._keep_track_on_concentration_to_calculation_mapping(a_str_id, i)

                    if a_str_id in self._concentrations_to_compounds_map:
                        max_concentrations = self._concentrations_to_compounds_map[a_str_id][0]
                        final_concentrations = self._concentrations_to_compounds_map[a_str_id][1]
                        concentration_fluxes = self._concentrations_to_compounds_map[a_str_id][2]
                    else:
                        max_concentrations = query_concentrations(
                            self.max_concentration_label, compound, self._properties, self._structures)
                        final_concentrations = query_concentrations(
                            self.final_concentration_label, compound, self._properties, self._structures)
                        concentration_fluxes = query_concentrations(
                            self.concentration_flux_label, compound, self._properties, self._structures)
                        self._concentrations_to_compounds_map[a_str_id] = (max_concentrations, final_concentrations,
                                                                           concentration_fluxes)
                    c_i = self._get_concentration_index_for_calculation(a_str_id, i)
                    concentrations = [max_concentrations[c_i], final_concentrations[c_i], concentration_fluxes[c_i]]
                    if None not in concentrations:
                        if concentrations[self._comparison_index] >= comparison_threshold:
                            label = str(round(concentrations[0], 4)) + "\n" + str(round(concentrations[1], 4))\
                                + "\n" + str(round(concentrations[2], 4))
                            compound_ids.append((c_id, label))
            self._compound_ids_per_calculations.append(compound_ids)
            self._old_threshold = comparison_threshold
        return self._compound_ids_per_calculations


class DirectedExplorationProgressSettings(QWidget):
    def __init__(self, parent: QWidget, db_manager: db.Manager) -> None:
        super(DirectedExplorationProgressSettings, self).__init__(parent)
        self._db_manager = db_manager
        self._comparison_label = "max_concentration"
        self._comparison_threshold = 1e-1
        self.show_only_added_compounds: bool = True
        self._label_options_cb: Union[None, BaseBox] = None
        layout = QVBoxLayout()
        self._width = 240
        self.setFixedWidth(self._width)

        self._button_update = QPushButton("Update")
        layout.addWidget(self._button_update)
        self._button_update.clicked.connect(self.__update_function)  # pylint: disable=no-member
        mapper_options = ["kinetic modeling"]
        self._step_to_compound_mapper = KineticModelingDrivenExplorationExtractor(self._db_manager)

        self._qvbox = QVBoxLayout()
        self._step_mapper_label = QLabel()
        self._step_mapper_label.setText("Exploration Step Mapper")
        self._qvbox.addWidget(self._step_mapper_label)

        self._mapper_cb = BaseBox()
        for mapper in mapper_options:
            self._mapper_cb.addItem(mapper)
        self._mapper_cb.currentIndexChanged.connect(self.update_step_to_compound_mapper)  # pylint: disable=no-member
        self._qvbox.addWidget(self._mapper_cb)

        self._concentration_box_label = QLabel()
        self._concentration_box_label.setText("Comparison Label")
        self._qvbox.addWidget(self._concentration_box_label)
        self.update_label_options(self._qvbox)

        self._concentration_threshold_label = QLabel()
        self._concentration_threshold_label.setText("Concentration Threshold")
        self._qvbox.addWidget(self._concentration_threshold_label)
        self._concentration_threshold_text = QLineEdit(self)
        self._concentration_threshold_text.setText(str(self._comparison_threshold))
        self._qvbox.addWidget(self._concentration_threshold_text)

        self.only_added_cbox = QCheckBox("Show only added compounds")
        self.only_added_cbox.setChecked(self.show_only_added_compounds)
        self.only_added_cbox.toggled.connect(self.update_show_only_added_compounds)  # pylint: disable=no-member
        self._qvbox.addWidget(self.only_added_cbox)

        self._qvbox.addStretch()

        layout.addLayout(self._qvbox)
        self.setLayout(layout)

    def update_show_only_added_compounds(self):
        self.show_only_added_compounds = self.only_added_cbox.isChecked()

    def update_label_options(self, layout: QVBoxLayout):
        label_options = self._step_to_compound_mapper.get_comparison_options()
        old_widget = self._label_options_cb
        self._label_options_cb = BaseBox()
        for o in label_options:
            self._label_options_cb.addItem(o)
        self._label_options_cb.currentIndexChanged.connect(self.update_comparison_index)  # pylint: disable=no-member
        if old_widget:
            layout.replaceWidget(old_widget, self._label_options_cb)
        else:
            layout.addWidget(self._label_options_cb)

    def __update_function(self) -> None:
        self.parent().progress_view.update_view(self._step_to_compound_mapper)

    def update_step_to_compound_mapper(self, new_mapper_index: int):
        if new_mapper_index == 0:
            self._step_to_compound_mapper = KineticModelingDrivenExplorationExtractor(self._db_manager)
            self.update_label_options(self._qvbox)

    def update_comparison_index(self, new_comparison_index):
        self._step_to_compound_mapper.set_comparison_index(new_comparison_index)

    def get_comparison_threshold(self):
        return float(self._concentration_threshold_text.text())


class KineticExplorationProgressWidget(QWidget):
    def __init__(self, parent: QWidget, db_manager: db.Manager) -> None:
        super(KineticExplorationProgressWidget, self).__init__(parent=parent)
        self.db_manager = db_manager

        # Create layout and add widgets
        layout = QHBoxLayout()
        self.progress_view = DirectedExplorationProgressView(self.db_manager)
        self.settings = DirectedExplorationProgressSettings(self, self.db_manager)
        layout.addWidget(self.progress_view)
        layout.addWidget(self.settings)

        # Set dialog layout
        self.setLayout(layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional, Dict
import pickle

from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QDialog,
    QLabel,
    QWidget,
    QLineEdit,
    QCompleter,
    QPushButton,
    QTextEdit,
    QCheckBox
)

from scine_utilities import ElementInfo

from scine_chemoton.reaction_rules import (
    valid_element,
    RuleSet,
    BaseRule
)

# we are using star imports here, because we allow for straight text save/load, which requires to know all classes
# in these submodules
from scine_chemoton.reaction_rules.\
     element_rules import *  # pylint: disable=(wildcard-import,unused-wildcard-import)  # noqa
from scine_chemoton.reaction_rules.\
    distance_rules import *  # pylint: disable=(wildcard-import,unused-wildcard-import)  # noqa
from scine_chemoton.reaction_rules.\
    polarization_rules import *  # pylint: disable=(wildcard-import,unused-wildcard-import)  # noqa
from scine_chemoton.reaction_rules.\
    reaction_rule_library import *  # pylint: disable=(wildcard-import,unused-wildcard-import)  # noqa

# explicitly import the classes we specify as types in the code to avoid pylint errors
from scine_chemoton.reaction_rules.distance_rules import (
    DistanceRuleSet,
    DistanceBaseRule,
)
from scine_chemoton.reaction_rules.element_rules import (
    ElementRuleSet,
    ElementBaseRule,
)
from scine_chemoton.reaction_rules.polarization_rules import (
    PolarizationRuleSet,
    PolarizationBaseRule,
)
from scine_chemoton.reaction_rules.reaction_rule_library import DefaultOrganicChemistry

from scine_heron.chemoton.class_selection_widget import ClassSelectionWidget
from scine_heron.containers.layouts import VerticalLayout, HorizontalLayout
from scine_heron.containers.buttons import TextPushButton
from scine_heron.io.text_box import yes_or_no_question
from scine_heron.io.file_browser_popup import get_save_file_name, get_load_file_name
from scine_heron.utilities import write_error_message, write_info_message
import scine_heron.io.json_pickle_wrap as json_wrap


class RuleBuilder(QDialog):
    """
    A pop-up that allows to build a Chemoton reaction rule.

    Notes
    -----
    This is mainly a wrapper around the ClassSelection widget
    """

    def __init__(self, rule_type: type, parent: Optional[QObject] = None):
        """
        Construct the widget with the allowed rule type.

        Parameters
        ----------
        rule_type : type
            The rule type we want to build
        parent : Optional[QObject], optional
            The parent widget, by default None

        Raises
        ------
        NotImplementedError
            Unknown rule type provided
        """
        super().__init__(parent=parent)
        if rule_type not in RuleSetBuilder.mapping_set_to_rule_type.values():
            raise NotImplementedError(f"Rule type {rule_type} not implemented")
        # widgets
        self._select_widget = ClassSelectionWidget(rule_type, "rule", parent=self)
        button = TextPushButton("Confirm", self.reject)
        # layout
        layout = VerticalLayout([self._select_widget, button])
        self.setLayout(layout)
        self.setWindowTitle("Rule Builder")

    def get_rule(self) -> BaseRule:
        """
        Get the rule instance that was built.

        Returns
        -------
        BaseRule
            The built rule

        Raises
        ------
        ValueError
            When the rule could not be constructed
        """
        rule = self._select_widget.get_instance()
        if rule is None:
            write_error_message("Could not construct a rule")
            raise ValueError("Could not construct a rule")
        return rule

    def get_current_rule_text(self) -> str:
        """
        Relies on `repr` on the current instance, empty string if no currently valid
        built text.

        Returns
        -------
        str
            The current rule text
        """
        rule = self._select_widget.get_instance()
        if rule is None:
            return ""
        return repr(rule)


class RuleSetBuilder(QDialog):
    """
    A pop-up that allows to build a set of Chemoton reaction rule

    Notes
    -----
    This is mainly a wrapper around the ClassSelection widget
    """

    # defines the connection between the RuleSet and Rule types
    mapping_set_to_rule_type: Dict[type, type] = {
        DistanceRuleSet: DistanceBaseRule,
        ElementRuleSet: ElementBaseRule,
        PolarizationRuleSet: PolarizationBaseRule,
    }

    def __init__(self, options: RuleSet, parent: Optional[QObject] = None):
        """
        Construct the widget with a set of rules that should be expanded

        Parameters
        ----------
        options : RuleSet
            The set to be expanded
        parent : Optional[QObject], optional
            The parent widget, by default None

        Raises
        ------
        NotImplementedError
            Unknown rule set type
        """
        rule_set_type = type(options)
        if rule_set_type not in self.mapping_set_to_rule_type:
            raise NotImplementedError(f"Unknown rule set type {rule_set_type}")
        super().__init__(parent=parent)
        self._options = options

        # rule builder
        self._rule_type = self.mapping_set_to_rule_type[rule_set_type]
        self._select_widget = ClassSelectionWidget(self._rule_type, "rule", parent=self)

        # load
        load_button = TextPushButton("Load existing rule set", self._load)
        if self._rule_type == DistanceBaseRule:
            default_button = TextPushButton("Load default organic chemistry", self._load_default_organic_chemistry)
        else:
            default_button = None

        # element line
        label = QLabel("Set rule for element:")
        self._element_edit = QLineEdit("C")
        completer = QCompleter([str(e) for e in ElementInfo.all_implemented_elements()])
        self._element_edit.setCompleter(completer)
        self._check_box = QCheckBox("move rule, when set")
        self._check_box.setChecked(True)
        button = TextPushButton("Set", self._add_rule, shortcut="Return")
        remove_button = TextPushButton("Remove", self._remove_rule)
        element_layout = HorizontalLayout([label, self._element_edit, self._check_box, button, remove_button])

        # display of all rules
        display_label = QLabel("Current rules:")
        self._display = QTextEdit()
        self._display.setReadOnly(True)
        details_button = TextPushButton("Show current rules in detail", self._detail_display, shortcut="Ctrl+S")

        # confirm
        close_button = TextPushButton("Confirm", self.reject)

        # total layout
        element_container = QWidget()
        element_container.setLayout(element_layout)
        order = [self._select_widget, load_button, default_button, element_container,
                 display_label, self._display, details_button, close_button]
        layout = VerticalLayout([widget for widget in order if widget is not None])
        self.setLayout(layout)
        self.setWindowTitle("Rule Set Builder")

    def get_current_rule_text(self) -> str:
        """
        Combines all rules in the current rule set into a text

        Returns
        -------
        str
            The text describing the set
        """
        text = ""
        for k, v in self._options.items():
            text += f"Rule for {k}: {repr(v)}\n"
        return text

    def _add_rule(self):
        """
        Add a single rule based on the input of the selection widget and the current element.
        """
        element = self._element_edit.text().strip()
        if not valid_element(element):
            write_error_message(f"Invalid element {element}, no rule added")
            return
        rule = self._select_widget.get_instance()
        if rule is None:
            write_error_message("Provided invalid rule, no rule added")
            return
        if element in self._options:
            answer = yes_or_no_question(self, f"Rule for element {element} already exists, overwrite")
            if not answer:
                write_info_message("No rule added")
                self._element_edit.setText("")
                return
        self._options[element] = rule
        if self._check_box.isChecked():
            self._select_widget.restart()  # remove written rule
        self._update_display()

    def _remove_rule(self):
        """
        Remove the current rule of the current element in the rule set
        """
        element = self._element_edit.text().strip()
        if element not in self._options:
            write_error_message(f"Haven't specified any rule for element {element}")
            return
        del self._options[element]
        self._update_display()

    def _update_display(self) -> None:
        """
        Updates our text display based on the current rule set
        """
        text = ""
        for k, v in self._options.items():
            text += f"Rule for {k}: {v.name}\n"
        self._display.setText(text)

    def _detail_display(self) -> None:
        """
        Constructs a pop-up widget with the full details of the current rules and
        includes a save button to write the rule set to disk.
        """
        popup = QDialog(self)
        # generate widgets
        label = QLabel("Detailed rule set")
        text = self.get_current_rule_text()
        text_box = QTextEdit()
        text_box.setText(text)  # two lines init to keep line breaks
        save = TextPushButton("Save rule set to disk", self._save)
        close = TextPushButton("Close", popup.reject)
        # fill layout
        layout = VerticalLayout([label, text_box, save, close])
        popup.setLayout(layout)
        popup.exec_()

    def _save(self) -> None:
        filename = get_save_file_name(self, f"{self._rule_type.__name__}_set",
                                      ["txt", "pkl", "pickle", "json"])
        if filename is None:
            return
        if filename.suffix == ".json":
            with open(filename, "w") as f:
                f.write(json_wrap.encode(self._options))
        elif filename.suffix == ".txt":
            # relies on that every rule set and rule has a proper __repr__
            with open(filename, "w") as f:
                f.write(repr(self._options))
        elif filename.suffix in [".pkl", "pickle"]:
            with open(filename, "wb") as f:
                pickle.dump(self._options, f)
        else:
            # DevNote: Forgot to implement type or file dialog not working
            write_error_message(f"Saving '{filename}' failed")

    def _load(self) -> None:
        filename = get_load_file_name(self, f"{self._rule_type.__name__}_set",
                                      ["txt", "pkl", "pickle", "json"])
        if filename is None:
            return
        if filename.suffix == ".json":
            with open(filename, "r") as f:
                data = f.read()
            loaded_options = json_wrap.decode(data)
        elif filename.suffix == ".txt":
            # relies on that every rule set and rule has a proper __repr__
            with open(filename, "r") as f:
                data = f.read()
            loaded_options = eval(data)  # pylint: disable=eval-used
        elif filename.suffix in [".pkl", "pickle"]:
            with open(filename, "rb") as f:
                loaded_options = pickle.load(f)
        else:
            # DevNote: Forgot to implement type or file dialog not working
            write_error_message(f"Loading '{filename}' failed")
            return
        for k, v in loaded_options.items():
            self._options[k] = v
        self._update_display()

    def _load_default_organic_chemistry(self) -> None:
        ruleset = DefaultOrganicChemistry()
        for k, v in ruleset.items():
            self._options[k] = v
        self._update_display()


class RuleSetBuilderButtonWrapper(QPushButton):

    def __init__(self, options: RuleSet, parent: Optional[QObject] = None):
        super().__init__(parent=parent)
        self.setText("Generate rule set")
        self._options = options
        self._builder = RuleSetBuilder(self._options, parent=self)
        self.clicked.connect(self.execute)  # pylint: disable=no-member
        self.setToolTip(self._builder.get_current_rule_text())

    def execute(self) -> None:
        self._builder.exec_()
        self.setToolTip(self._builder.get_current_rule_text())

    def get_options(self) -> RuleSet:
        return self._options


class RuleBuilderButtonWrapper(QPushButton):

    def __init__(self, rule_type: type, parent: Optional[QObject] = None):
        super().__init__(parent=parent)
        self.setText("Generate rule")
        self._builder = RuleBuilder(rule_type, parent=self)
        self.clicked.connect(self.execute)  # pylint: disable=no-member
        self.setToolTip(self._builder.get_current_rule_text())

    def execute(self) -> None:
        self._builder.exec_()
        self.setToolTip(self._builder.get_current_rule_text())

    def get_rule(self) -> BaseRule:
        return self._builder.get_rule()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from abc import ABC, abstractmethod
from copy import deepcopy
from time import sleep
from typing import Any, List, Optional, Union, TYPE_CHECKING, Type, Dict

from PySide2.QtCore import QThread
from PySide2.QtWidgets import QWidget

if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


import scine_database as db
import scine_utilities as su
import scine_swoose as swoose  # noqa # pylint: disable=import-error

from scine_heron.molecular_viewer import MolecularViewerWidget, get_mol_viewer_tab


class _SwooseWrapper(ABC):

    def __init__(self) -> None:
        self.errors: List[str] = []
        self._db_manager: Optional[db.Manager] = None
        self._mol_tab: Optional[MolecularViewerWidget] = None
        self._instance: Union[swoose.QmRegionSelector, swoose.Parametrizer, None] = None

    def _get_excluded_settings(self) -> List[str]:
        excluded_settings = [
            "qm_region_center_atoms",
            "ref_data_mode",
            "yaml_settings_file_path",
        ]
        if not self.db_available():
            excluded_settings.extend([
                "database_host",
                "database_name",
                "database_port",
                "database_sleep_time",
                "reuse_database",
            ])
        return excluded_settings

    def _get_default_modified_settings(self) -> Dict[str, Any]:
        if not self.db_available():
            return {}
        assert self._db_manager is not None  # just linting, proper check in method
        credentials = self._db_manager.get_credentials()
        return {
            "database_host": credentials.hostname,
            "database_name": credentials.database_name,
            "database_port": credentials.port
        }

    def identifier(self) -> str:
        return self._instance.__class__.__name__

    @abstractmethod
    def failed_result(self) -> Union[List[int], bool]:
        pass

    @abstractmethod
    def sanity_check(self) -> Union[List[int], bool]:
        pass

    @abstractmethod
    def task(self) -> Union[List[int], bool]:
        pass

    def get_settings(self) -> su.Settings:
        if self._instance is None:
            return su.Settings("none", {})
        shown_settings = deepcopy(self._instance.settings)
        for excluded in self._get_excluded_settings():
            if excluded in shown_settings:
                del shown_settings[excluded]
        for key, value in self._get_default_modified_settings().items():
            shown_settings[key] = value
        return shown_settings

    def set_settings(self, settings: su.Settings) -> None:
        if self._instance is None:
            return
        self._instance.settings.update(settings)

    def _reload_mol_tab(self):
        self._mol_tab = self._mol_view_wrap()

    def _mol_view_wrap(self) -> Optional[MolecularViewerWidget]:
        tab = get_mol_viewer_tab(want_atoms_there=True, message_container=self.errors)
        return tab

    def enable_database(self, db_manager: db.Manager) -> None:
        self._db_manager = db_manager
        if not db_manager.connected:
            try:
                self._db_manager.connect()
            except RuntimeError as e:
                self.errors.append(str(e))

    def disable_database(self) -> None:
        self._db_manager = None

    def db_available(self) -> bool:
        return self._db_manager is not None and self._db_manager.connected


class SwooseTask(QThread):

    # signals that this threads sends to main display for result of Swoose Task
    parametrize_signal = Signal(bool)
    qm_region_signal = Signal(list)

    # technical reasons
    error_signal = Signal(str)  # error message cannot be displayed from thread
    send_settings_signal = Signal(su.Settings, str)  # pop-up for settings cannot safely be displayed from thread
    receive_settings_signal = Signal(su.Settings)

    def __init__(self, parent: Optional[QWidget], swoose_wrapper: Type):
        super().__init__(parent)
        self._swoose_wrapper: _SwooseWrapper = swoose_wrapper()
        self.identifier = self._swoose_wrapper.identifier()
        self._settings: Optional[su.Settings] = None

    def enable_database(self, db_manager: db.Manager) -> None:
        self._swoose_wrapper.enable_database(db_manager)
        self._error_check()

    def disable_database(self) -> None:
        self._swoose_wrapper.disable_database()

    def _error_check(self, additional_error: str = "") -> None:
        if self._swoose_wrapper.errors:
            for error in self._swoose_wrapper.errors:
                self.error_signal.emit(error)
        if additional_error:
            self.error_signal.emit(additional_error)
            self._send_result(self._swoose_wrapper.failed_result())

    def _send_result(self, result: Union[List[int], bool]) -> None:
        if isinstance(result, list):
            self.qm_region_signal.emit(result)
        elif isinstance(result, bool):
            self.parametrize_signal.emit(result)
        else:
            raise NotImplementedError(f"Unknown Swoose result {result}")

    def run(self):
        try:
            sane_result = self._swoose_wrapper.sanity_check()
            if not sane_result:
                self._error_check()
                self._send_result(self._swoose_wrapper.failed_result())
                self.exit(1)
            self._wait_for_settings()
            if self._settings is None:
                self._error_check(additional_error="Could not get settings")
                self.exit(0)
            self._swoose_wrapper.set_settings(self._settings)
            result = self._swoose_wrapper.task()
        except Exception as e:  # pylint: disable=broad-except
            self._error_check(additional_error=str(e))
        else:
            self._error_check()
            self._send_result(result)
        self.exit(0)

    def _stop_waiting(self, settings: su.Settings) -> None:
        self._settings = settings
        self.receive_settings_signal.disconnect(self._stop_waiting)

    def _wait_for_settings(self) -> None:
        settings = self._swoose_wrapper.get_settings()
        self.receive_settings_signal.connect(self._stop_waiting)
        self.send_settings_signal.emit(settings, self.identifier)
        while self._settings is None:
            sleep(0.1)


class QMRegionSelector(_SwooseWrapper):

    def __init__(self):
        super().__init__()
        self._instance = swoose.QmRegionSelector()

    def _get_excluded_settings(self) -> List[str]:
        return super()._get_excluded_settings() + ["qm_region_center_atoms"]

    def _get_default_modified_settings(self) -> Dict[str, Any]:
        return {**super()._get_default_modified_settings(),
                **{
                    "cutting_probability": 0.6,
                    "qm_region_min_size": 40,
                    "qm_region_max_size": 60,
                    "ref_max_size": 80,
                    "tol_percentage_sym_score": 99.0
        }}

    def failed_result(self) -> List[int]:
        return []

    def sanity_check(self) -> Union[List[int], bool]:
        self._reload_mol_tab()
        selection = self._get_mol_tab_atom_selection()
        if not selection:
            return []
        return True

    def task(self) -> Union[List[int], bool]:
        self._reload_mol_tab()
        selection = self._get_mol_tab_atom_selection()
        if not selection:
            return []
        self._set_hardcoded_settings(selection)
        tab = get_mol_viewer_tab(True)
        assert isinstance(tab, MolecularViewerWidget)
        assert tab.mol_widget is not None
        atoms = tab.mol_widget.get_atom_collection()
        calculator = tab.create_calculator_widget.get_calculator()
        calculator.structure = atoms
        assert calculator.name() == "QMMM"
        assert isinstance(self._instance, swoose.QmRegionSelector) and self._mol_tab is not None \
            and self._mol_tab.mol_widget is not None
        self._instance.set_underlying_calculator(calculator)
        self._instance.generate_qm_region(self._mol_tab.mol_widget.get_atom_collection())
        return self._instance.get_qm_region_indices()

    def _get_mol_tab_atom_selection(self) -> List[int]:
        if self._mol_tab is None:
            self.errors.append("Cannot construct QM region")
            return []
        mol_widget = self._mol_tab.mol_widget
        if mol_widget is None or not mol_widget.has_atoms():
            self.errors.append("Cannot construct QM region")
            return []
        selection = mol_widget.get_selection()
        if not selection:
            self.errors.append("No selection, please select one nucleus to start the QM region construction "
                               "around this atom")
            return []
        return selection

    def _set_hardcoded_settings(self, selection: List[int]) -> None:
        if self._instance is None:
            return
        mode = "database" if self.db_available() else "direct"
        self._instance.settings["ref_data_mode"] = mode
        self._instance.settings["qm_region_center_atoms"] = selection


class MMParametrizer(_SwooseWrapper):

    def __init__(self):
        super().__init__()
        self._instance = swoose.Parametrizer()

    def _get_excluded_settings(self) -> List[str]:
        return super()._get_excluded_settings() + ["ref_data_generation_only"]

    def _get_default_modified_settings(self) -> Dict[str, Any]:
        return {**super()._get_default_modified_settings(),
                **{
                    "reference_program": "sparrow",
                    "mm_parameter_file": "Parameters.dat",
                    "mm_connectivity_file": "Connectivity.dat",
        }}

    def failed_result(self) -> bool:
        return False

    def sanity_check(self) -> Union[List[int], bool]:
        self._reload_mol_tab()
        if self._mol_tab is None:
            self.errors.append("Cannot parametrize")
            return False
        mol_widget = self._mol_tab.mol_widget
        if mol_widget is None or not mol_widget.has_atoms():
            self.errors.append("Cannot parametrize")
            return False
        return True

    def task(self) -> Union[List[int], bool]:
        self._reload_mol_tab()
        if self._mol_tab is None:
            self.errors.append("Cannot parametrize")
            return False
        mol_widget = self._mol_tab.mol_widget
        if mol_widget is None or not mol_widget.has_atoms():
            self.errors.append("Cannot parametrize")
            return False
        self._set_hardcoded_settings()
        assert isinstance(self._instance, swoose.Parametrizer)
        self._instance.parametrize_mm(mol_widget.get_atom_collection())
        return True

    def _set_hardcoded_settings(self) -> None:
        if self._instance is None:
            return
        mode = "database" if self.db_available() else "direct"
        self._instance.settings["ref_data_mode"] = mode



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional

from PySide2.QtWidgets import QFileDialog, QPushButton, QVBoxLayout, QWidget

# from scine_heron.autocas.autocas_settings import AutocasSettings
# from scine_heron.autocas.signal_handler import SignalHandler
# from scine_heron.electronic_data.molden_file_reader import MoldenFileReader
from scine_heron.molecule.molecule_widget import MoleculeWidget
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.status_manager import StatusManager

from pathlib import Path


class OrbitalViewerWidget(QWidget):
    def __init__(self, parent, signal_handler, autocas_settings):
        QWidget.__init__(self, parent)
        self.__layout = QVBoxLayout()

        self.signal_handler = signal_handler
        self.autocas_settings = autocas_settings
        self.signal_handler.load_xyz_file_signal.connect(self.__load_molecule)
        self.signal_handler.load_molden_file_signal.connect(self.__load_orbitals)
        self.signal_handler.view_orbital.connect(self.__view_orbital)

        self.molecule_widget = QWidget(self)
        self.settings_status_manager = SettingsStatusManager()  # Settings())
        self.__enabled = StatusManager(True)

        self.__layout.addWidget(self.molecule_widget)

        self.button_load_molecule = QPushButton("Load Molecule")
        self.__layout.addWidget(self.button_load_molecule)
        # pylint: disable-next=E1101
        self.button_load_molecule.clicked.connect(self.__load_molecule)

        self.orbital_file = ""
        self.button_load_orbitals = QPushButton("Load Orbitals")
        self.__layout.addWidget(self.button_load_orbitals)
        # pylint: disable-next=E1101
        self.button_load_orbitals.clicked.connect(self.__load_orbitals)

        self.setLayout(self.__layout)

    def set_enabled(self, enabled: bool) -> None:
        """
        Set widgets enabled.
        """
        self.__enabled.value = enabled

    def __load_molecule(self, xyz_file: str = ""):
        # print("initial XYZ File", xyz_file)
        if xyz_file:
            self.autocas_settings.molecule_xyz_file = xyz_file
        else:
            self.autocas_settings.molecule_xyz_file, _ = QFileDialog.getOpenFileName(
                self,
                self.tr("Open XYZ File"),  # type: ignore[arg-type]
                "",
                self.tr("Molecule (*.xyz)"),  # type: ignore[arg-type]
            )
            # print("Loading XYZ file")
            # print(self.autocas_settings.molecule_xyz_file)
        old_widget = self.molecule_widget
        self.molecule_widget = MoleculeWidget(
            settings_status_manager=self.settings_status_manager,
            parent=self,
            disable_modification=True,
            file=Path(self.autocas_settings.molecule_xyz_file),
        )

        self.layout().replaceWidget(old_widget, self.molecule_widget)

    def __load_orbitals(self, orbital_file: Optional[str] = None):
        if not isinstance(self.molecule_widget, MoleculeWidget):
            # print("No molecular structure available. Loading now!")
            self.__load_molecule(self.autocas_settings.molecule_xyz_file)
        if orbital_file:
            self.orbital_file = orbital_file
        else:
            self.orbital_file, _ = QFileDialog.getOpenFileName(
                self,
                self.tr("Open Orbital File"),  # type: ignore[arg-type]
                "",
                self.tr("Molden file (*.molden)"),  # type: ignore[arg-type]
            )
            # print("Loading Molden file")
            # print(self.orbital_file)
        self.autocas_settings.molcas_orbital_file = self.orbital_file
        with open(self.orbital_file) as f:
            lines = f.read()

        self.molecule_widget.get_electronic_data_widget().read_molden_input(lines)
        self.autocas_settings.orbital_energies = self.molecule_widget.get_electronic_data_widget()\
            .get_electronic_data().get_orbital_energies()
        self.autocas_settings.orbital_occupations = self.molecule_widget.get_electronic_data_widget()\
            .get_electronic_data().get_occupations()
        self.signal_handler.update_mo_diagram.emit()
        self.molecule_widget.get_electronic_data_widget().view_orbital(5)

    def __view_orbital(self, index):
        self.molecule_widget.get_electronic_data_widget().view_orbital(index)

        # self.settings_status_manager.selected_mo_changed.emit(6)
        # self.molecule_widget.update()
        # self.settings_status_manager.selected_molecular_orbital = 4
        # self.__enabled.changed_signal.connect(self.button_load_orbitals.setEnabled)
        self.molecule_widget.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from collections import defaultdict
from threading import Event
from typing import Optional, List, Any, Dict, Tuple, TYPE_CHECKING

import numpy as np
from PySide2.QtCore import QThread
from PySide2.QtWidgets import QWidget

from scine_database import Manager, Credentials, Compound, Structure
from scine_utilities import AtomCollection
from scine_chemoton.steering_wheel.datastructures import ProtocolEntry, SelectionResult
from scine_chemoton.steering_wheel.network_expansions import (
    NetworkExpansion,
    GiveWholeDatabaseWithModelResult
)
from scine_chemoton.gears.conformers.brute_force import BruteForceConformers
from scine_chemoton.gears.elementary_steps import ElementaryStepGear

if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class UnimolecularCoordinates:
    """
    This class mitigates any problems when sending the coordinates as a signal with optional types and defaultdict.
    """

    def __init__(self, data: Dict[str,
                                  Dict[str,
                                       Optional[List[Tuple[List[List[Tuple[int, int]]], int]]]]]) -> None:
        self.data = data


class BimolecularCoordinates:
    """
    This class mitigates any problems when sending the coordinates as a signal with optional types and defaultdict.
    """

    def __init__(self, data: Dict[str,
                                  Dict[str,
                                       Dict[Tuple[List[Tuple[int, int]], int],
                                            List[Tuple[np.ndarray, np.ndarray, float, float]]
                                            ]
                                       ]
                                  ]) -> None:
        self.data = data


class CoordinateThread(QThread):  # no ABC because of QThread

    info_message_signal = Signal(str)
    error_message_signal = Signal(str)
    loop_signal = Signal()
    final_count_signal = Signal(int)

    def __init__(self, parent: QWidget, credentials: Credentials, potential_next_step: NetworkExpansion,
                 selection: SelectionResult) -> None:
        super().__init__(parent=parent)
        self._manager = Manager()
        self._credentials = credentials
        self._selection = selection
        self._manager.set_credentials(credentials)
        self._manager.connect()
        self._potential_next_step = potential_next_step
        self._potential_next_step.protocol = []
        self._compound_collection = self._manager.get_collection("compounds")
        self._structure_collection = self._manager.get_collection("structures")
        self._was_stopped = Event()

    def _setup_step(self) -> None:
        self._potential_next_step.dry_setup_protocol(self._credentials, self._selection)

    def _cleanup(self):
        self._potential_next_step.protocol = []
        if self._manager is not None:
            self._manager.disconnect()
            self._manager = None

    def run(self):
        try:
            self._setup_step()
            self._run_impl()
        except BaseException as e:
            self.error_message_signal.emit(str(e))  # pylint: disable=no-member
            self.final_count_signal.emit(0)  # pylint: disable=no-member
        finally:
            self._cleanup()
            self.exit(0)

    def _run_impl(self) -> None:
        raise NotImplementedError("This method needs to be implemented in a subclass")

    def terminate(self):
        for p in self._potential_next_step.protocol:
            if not isinstance(p, ProtocolEntry):
                continue
            p.terminate()
        self._cleanup()
        super().terminate()

    def stop(self):
        self._was_stopped.set()
        for p in self._potential_next_step.protocol:
            if not isinstance(p, ProtocolEntry):
                continue
            p.engine.stop()
        for p in self._potential_next_step.protocol:
            if not isinstance(p, ProtocolEntry):
                continue
            p.engine.join()


class UnimolecularThread(CoordinateThread):
    """
    """

    single_structure_no_sites_signal = Signal(str, AtomCollection)
    coordinates_signal = Signal(UnimolecularCoordinates)

    def _run_impl(self) -> None:
        """
        Queries for the unimolecular coordinates and emits the results.
        """
        total: Dict[str,
                    Dict[str,
                         Optional[List[Tuple[List[List[Tuple[int, int]]], int]]]]] \
            = defaultdict(lambda: defaultdict(list))
        if isinstance(self._potential_next_step, GiveWholeDatabaseWithModelResult):
            # loop over db
            model = self._potential_next_step.options.model
            for compound in self._compound_collection.iterate_all_compounds():
                self.loop_signal.emit()  # pylint: disable=no-member
                compound.link(self._compound_collection)
                if self._was_stopped.is_set():
                    break
                for sid in compound.get_structures():
                    structure = Structure(sid, self._structure_collection)
                    if structure.get_model() == model:
                        self.single_structure_no_sites_signal.emit(  # pylint: disable=no-member
                            f"Compound ({str(sid)})", structure.get_atoms()
                        )
                        break
            self.final_count_signal.emit(0)  # pylint: disable=no-member
            return
        for entry in self._potential_next_step.protocol:
            if not isinstance(entry, ProtocolEntry):
                continue
            if isinstance(entry.gear, ElementaryStepGear):
                entry.gear.disable_caching()
                uni_result = entry.gear.unimolecular_coordinates(
                    self._credentials, self.loop_signal.emit  # pylint: disable=no-member
                )
                if self._was_stopped.is_set():
                    break
                for cid, sub_dict in uni_result.items():
                    for sid, coords in sub_dict.items():
                        if sid in total[cid]:
                            total[cid][sid] += coords
                        else:
                            total[cid][sid] = coords
            elif isinstance(entry.gear, BruteForceConformers):
                if total:
                    self.info_message_signal.emit(  # pylint: disable=no-member
                        "Cannot show conformer generation + reaction trials, only showing conformer generation"
                    )
                entry.gear.clear_cache()
                count = 0
                for cid in entry.gear.valid_compounds():
                    self.loop_signal.emit()  # pylint: disable=no-member
                    compound = Compound(cid, self._compound_collection)
                    centroid = Structure(compound.get_centroid(), self._structure_collection)
                    self.single_structure_no_sites_signal.emit(f"Compound ({str(cid)})",  # pylint: disable=no-member
                                                               centroid.get_atoms())
                    count += 1
                self.final_count_signal.emit(count)  # pylint: disable=no-member
                return
        if self._was_stopped.is_set():
            self.final_count_signal.emit(0)  # pylint: disable=no-member
        else:
            self.coordinates_signal.emit(UnimolecularCoordinates(total))  # pylint: disable=no-member
        self._cleanup()
        self.exit(0)


class BimolecularThread(CoordinateThread):

    coordinates_signal = Signal(BimolecularCoordinates)

    def _run_impl(self) -> None:
        """
        Queries for the bimolecular coordinates and emits the results.
        """

        total: Dict[str,
                    Dict[str,
                         Dict[Tuple[List[Tuple[int, int]], int],
                              List[Tuple[np.ndarray, np.ndarray, float, float]]
                              ]
                         ]
                    ] \
            = defaultdict(lambda: defaultdict(dict))
        for entry in self._potential_next_step.protocol:
            if not isinstance(entry, ProtocolEntry):
                continue
            if isinstance(entry.gear, ElementaryStepGear):
                entry.gear.disable_caching()
                bi_result = entry.gear.bimolecular_coordinates(
                    self._credentials, self.loop_signal.emit  # pylint: disable=no-member
                )
                for cid, sub_dict in bi_result.items():
                    for sid, coords in sub_dict.items():
                        for key, complexes in coords.items():
                            if key in total[cid][sid]:
                                total[cid][sid][key] += complexes
                            else:
                                total[cid][sid][key] = complexes
        if self._was_stopped.is_set():
            self.final_count_signal.emit(0)  # pylint: disable=no-member
        else:
            self.coordinates_signal.emit(BimolecularCoordinates(total))  # pylint: disable=no-member
        self._cleanup()
        self.exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import List, TypeVar, Optional, Callable

import numpy as np
from vtk import vtkMolecule

from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.edit_molecule.new_atom_positions import new_atom_position_from_id_list

Position = TypeVar("Position")


def _add_atom_to_molecule(molecule: vtkMolecule, N: int, position: Position) -> None:

    atom = molecule.AppendAtom()
    atom.SetAtomicNumber(N)
    atom.SetPosition(position)


def build_molecule_without_removed_atoms(
    old_molecule: vtkMolecule, id_atoms_to_remove: List[int]
) -> vtkMolecule:
    """
    Creates a new molecule copying all the atoms
    except the one with the given id.
    """
    new_molecule = vtkMolecule()
    for i in range(old_molecule.GetNumberOfAtoms()):
        if i not in id_atoms_to_remove:
            old_atom = old_molecule.GetAtom(i)
            _add_atom_to_molecule(
                new_molecule, old_atom.GetAtomicNumber(), old_atom.GetPosition()
            )

    return new_molecule


def build_molecule_with_new_atom_structural_completion(
    old_molecule: vtkMolecule,
    new_atomic_number: int,
    other_atom_ids: List[int],
    settings_status_manager: SettingsStatusManager,
) -> vtkMolecule:
    def structural_completion_strategy(
        molecule: vtkMolecule, new_n: int
    ) -> Optional[np.ndarray]:
        new_position, method = new_atom_position_from_id_list(
            other_atom_ids, new_n, molecule
        )
        settings_status_manager.info_message = "method used: " + str(method)
        return new_position

    return _build_molecule_with_new_atom(
        old_molecule, new_atomic_number, structural_completion_strategy
    )


def _build_molecule_with_new_atom(
    old_molecule: vtkMolecule,
    new_atomic_number: int,
    positioning_strategy: Callable[[vtkMolecule, int], Position],
) -> vtkMolecule:

    new_molecule = vtkMolecule()
    new_molecule.DeepCopy(old_molecule)

    new_position = positioning_strategy(new_molecule, new_atomic_number)

    if new_position is None:
        return new_molecule

    _add_atom_to_molecule(new_molecule, new_atomic_number, new_position)

    return new_molecule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the `create_molecule_animator` function.
"""
from scine_heron.status_manager import StatusManager
from scine_heron.haptic.haptic_client import HapticClient
from scine_heron.electronic_data.electronic_data_status_manager import (
    ElectronicDataStatusManager,
)
from scine_heron.energy_profile.energy_profile_point import EnergyProfilePoint
from scine_heron.energy_profile.energy_profile_status_manager import (
    EnergyProfileStatusManager,
)
from scine_heron.molecule.utils.molecule_utils import (
    molecule_to_list_of_atoms,
    convert_gradients,
    apply_gradients,
)
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.molecule.animator_pooling_functions import (
    calculate_gradient,
    GradientCalculationResult,
)
from scine_heron.molecule.animator import Animator
from typing import Optional, Tuple, List, Any, Dict
from vtk import vtkMolecule


def create_molecule_animator(
    molecule_version: int,
    molecule: vtkMolecule,
    settings_status_manager: SettingsStatusManager,
    haptic_client: Optional[HapticClient],
    energy_status_manager: Optional[EnergyProfileStatusManager],
    electronic_data_status_manager: Optional[ElectronicDataStatusManager],
    charge_status_manager: StatusManager[Optional[List[float]]],
    force_status_manager: StatusManager[Optional[List[float]]],
) -> Animator:
    """
    Creates an Animator that calculates and applies gradients to the molecule
    on every update.
    """

    def provide_data() -> Tuple[
        int, List[Tuple[str, Tuple[float, float, float]]], Tuple[str, str], Dict[str, Any], bool, str,
    ]:
        return (
            molecule_version,
            molecule_to_list_of_atoms(molecule),
            settings_status_manager.get_calculator_args(),
            settings_status_manager.get_calculator_settings(),
            settings_status_manager.get_mediator_potential_setting(),
            settings_status_manager.bond_display,
        )

    def apply_results(result: GradientCalculationResult, time_interval: float) -> None:
        charge_status_manager.value = result.atomic_charges
        force_status_manager.value = result.gradients.tolist()

        if energy_status_manager is not None:
            if len(energy_status_manager) == 0:
                if result.energy != 0.0:
                    elapsed_time = time_interval
                    energy_profile_point = EnergyProfilePoint(
                        result.energy, elapsed_time
                    )
                    energy_status_manager.append(energy_profile_point)
            else:
                if result.energy != 0.0:
                    energy = result.energy
                    energy_profile_point = EnergyProfilePoint(
                        energy, 0.0
                    )
                    previous_time = energy_status_manager.value[-1].elapsed_time
                    delta = energy_profile_point.time_stamp - energy_status_manager.value[-1].time_stamp
                    delta_float = delta.total_seconds()
                    energy_profile_point.elapsed_time = previous_time + delta_float

                    energy_status_manager.append(energy_profile_point)

        settings_status_manager.error_message = result.error_msg
        settings_status_manager.info_message = result.info_msg

        if electronic_data_status_manager is not None:
            electronic_data_status_manager.molden_input = result.molden_input
        convert_gradients(result.gradients, boost_factor=settings_status_manager.gradients_scaling)
        apply_gradients(
            molecule,
            result.gradients,
            result.bond_orders,
            settings_status_manager.mouse_picked_atom_ids,
            settings_status_manager.haptic_picked_atom_id,
        )
        if haptic_client is not None:
            haptic_client.update_molecule(molecule)
            haptic_client.update_gradient(result.gradients, settings_status_manager.haptic_force_scaling)

    return Animator(calculate_gradient, provide_data, apply_results)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Dict, Optional, Any, List, Union
import json
import numpy as np

from scine_database.queries import optimized_labels_enums
from scine_database.energy_query_functions import get_energy_for_structure
import scine_database as db
import scine_utilities as utils

from scine_heron.utilities import (
    color_axis,
    color_figure,
    get_font,
    get_primary_line_color,
    qcolor_by_key,
    build_brush,
    build_pen
)
from scine_heron.database.graphics_items import Compound, Structure, Reaction
from scine_heron.database.compound_and_flasks_helper import get_compound_or_flask
from scine_heron.molecule.molecule_widget import MoleculeWidget
from scine_heron.molecule.molecule_video import MoleculeVideo
from scine_heron.utilities import copy_text_to_clipboard
from scine_heron import get_core_tab

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure

from PySide2.QtWidgets import (
    QWidget,
    QGraphicsView,
    QGraphicsItem,
    QGraphicsScene,
    QVBoxLayout,
    QHBoxLayout,
    QMenu,
)
from PySide2.QtGui import (
    QPainterPath,
    QGuiApplication,
    QKeySequence,
    QPainter
)
from PySide2.QtCore import (
    Qt,
    QPoint,
    QTimeLine,
    QEvent,
)


class ExpandCompound(QWidget):

    def __init__(self, parent: Optional[QWidget], db_manager: db.Manager,
                 selected_aggregate: Dict[str, Any]) -> None:
        super(ExpandCompound, self).__init__(parent=parent)

        self.__selected_aggregate = selected_aggregate
        self.__structures = db_manager.get_collection('structures')
        self.__properties = db_manager.get_collection('properties')

        self.__molecular_viewer = MoleculeWidget(
            parent=self,
            alternative_zoom_controls=True,
            disable_modification=True
        )
        self.__molecular_viewer.setMinimumHeight(200)
        self.__molecular_viewer.setMinimumWidth(200)
        self.__figure = Figure(figsize=(3, 6))
        self.__plot_widget = FigureCanvasQTAgg(self.__figure)
        self.__plot_widget.setMinimumHeight(200)
        self.__plot_widget.setMinimumWidth(200)
        self.__network_widget = ExpandedSingleCompoundNetwork(self, db_manager, selected_aggregate)
        self.__network_widget.setMinimumHeight(400)
        self.__network_widget.setMinimumWidth(400)

        vbox_widget = QWidget()
        vbox = QVBoxLayout()
        vbox.addWidget(self.__molecular_viewer)
        vbox.addWidget(self.__plot_widget)
        vbox_widget.setLayout(vbox)
        self.__layout = QHBoxLayout()
        self.__layout.addWidget(self.__network_widget)
        self.__layout.addWidget(vbox_widget)
        self.setLayout(self.__layout)

        self.__none_duplicate_structures: List[str] = []
        self.__structure_energies_by_id: Dict[str, Optional[float]] = {}
        self.__min_energy: float = 0.0
        self.__number_of_highlights: int = 0
        self.__gather_data()
        self.__init_plot()
        self.__update_plot()

    def __gather_data(self):
        allowed_labels = optimized_labels_enums()
        for sid in [x['$oid'] for x in self.__selected_aggregate['structures']]:
            structure = db.Structure(db.ID(sid))
            structure.link(self.__structures)
            # TODO in the future users should have control over this model
            model = structure.get_model()
            if structure.get_label() not in allowed_labels:
                continue
            self.__none_duplicate_structures.append(sid)
            energy = get_energy_for_structure(
                structure,
                'electronic_energy',
                model,
                self.__structures,
                self.__properties
            )
            self.__structure_energies_by_id[sid] = energy
        self.__min_energy = min(self.__structure_energies_by_id.values())

    def __init_plot(self):
        self.__ax = self.__figure.add_subplot(1, 1, 1)
        color_figure(self.__figure)
        font = get_font()
        color_axis(self.__ax)
        self.__ax.set_title("Conformer Distribution", font)
        self.__ax.set_ylabel("Relative Electronic Energy in kJ/mol", font)
        self.__ax.set_xlabel(
            f"Min. Conformer Energy: {self.__min_energy:.5f} Hartree"
        )
        self.__ax.set_xticks([])
        self.__ax.xaxis.set_ticklabels([])
        # self.__ax.axes.xaxis.set_visible(False)
        x = [1.0, 2.0]
        for sid in self.__none_duplicate_structures:
            energy = self.__structure_energies_by_id[sid]
            if energy is None:
                continue
            rel_energy = (energy - self.__min_energy) * 2625.5
            y = [rel_energy, rel_energy]
            self.__ax.plot(x, y, color=get_primary_line_color())
        self.__plot_widget.draw()

    def __update_plot(self, highlights: Optional[List[str]] = None):
        x = [1.0, 2.0]
        for _ in range(self.__number_of_highlights):
            self.__ax.lines[-1].remove()
        self.__number_of_highlights = 0
        if highlights is None:
            return
        for sid in highlights:
            energy = self.__structure_energies_by_id[sid]
            if energy is None:
                continue
            rel_energy = (energy - self.__min_energy) * 2625.5
            y = [rel_energy, rel_energy]
            self.__ax.plot(x, y, color="#D40000", linewidth=2)
            self.__number_of_highlights += 1
        self.__plot_widget.draw()

    def update_plot_only(self, structure_id: Optional[str], additional_highlights: Optional[List[str]] = None):
        if structure_id is None or structure_id not in self.__none_duplicate_structures:
            self.__update_plot(highlights=additional_highlights)
        else:
            if additional_highlights is not None:
                self.__update_plot(highlights=[structure_id] + additional_highlights)
            else:
                self.__update_plot(highlights=[structure_id])

    def update_selection_dependent_widgets(
        self,
        structure_id: Optional[str],
        additional_highlights: Optional[List[str]] = None
    ):
        if structure_id is None or structure_id not in self.__none_duplicate_structures:
            atoms = utils.AtomCollection()
            self.__molecular_viewer.update_molecule(atoms=atoms)
            self.__update_plot(highlights=additional_highlights)
        else:
            structure = db.Structure(db.ID(structure_id))
            structure.link(self.__structures)
            atoms = structure.get_atoms()
            self.__molecular_viewer.update_molecule(atoms=atoms)
            if additional_highlights is not None:
                self.__update_plot(highlights=[structure_id] + additional_highlights)
            else:
                self.__update_plot(highlights=[structure_id])


class ExpandedSingleCompoundNetwork(QGraphicsView):
    def __init__(self, parent: Optional[QWidget], db_manager: db.Manager,
                 selected_aggregate: dict, unique_structures: Optional[List[str]] = None) -> None:
        self.scene_object = QGraphicsScene()
        super(ExpandedSingleCompoundNetwork, self).__init__(self.scene_object, parent=parent)
        self.setMouseTracking(True)
        self.setInteractive(True)

        # settings regarding presentation
        self.compound_color = qcolor_by_key('compoundColor')
        self.highlight_color = qcolor_by_key('highlightColor')
        self.border_color = qcolor_by_key('borderColor')
        self.edge_color = qcolor_by_key('edgeColor')
        self.structure_color = qcolor_by_key('structureColor')

        self.structure_brush = build_brush(self.structure_color)
        self.structure_pen = build_pen(self.border_color)
        self.compound_brush = build_brush(self.compound_color)
        self.compound_pen = build_pen(self.border_color)
        self.hover_brush = build_brush(self.highlight_color)
        self.hover_pen = build_pen(self.highlight_color, width=2)
        self.path_pen = build_pen(self.edge_color)

        # rendering smoother lines and edges of nodes
        self.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)

        self.__unique_structures = unique_structures
        if not self.__unique_structures:
            self.__unique_structures = []
            structure_collection = db_manager.get_collection("structures")
            allowed_labels = optimized_labels_enums()
            for sid in [x['$oid'] for x in selected_aggregate['structures']]:
                structure = db.Structure(db.ID(sid))
                structure.link(structure_collection)
                if structure.get_label() not in allowed_labels:
                    continue
                self.__unique_structures.append(sid)

        # Add all data
        self.db_manager: db.Manager = db_manager
        self.compounds: Dict[Union[str, int], Compound] = {}
        self.centroid_item: Optional[Compound] = None
        self.__current_focus: Optional[str] = None

        self.selected_compound: dict = selected_aggregate
        self.current_hover: Optional[str] = None
        self.item_list: List[List[db.ID]] = [[], []]

        self.plot_network_compound()

    def plot_network_compound(self) -> None:
        # Create singular timer for all added animated items
        self.timer = QTimeLine(1000)
        self.frames = 40
        self.timer.setFrameRange(0, self.frames)

        # db-info of selected compound
        compound_collection = self.db_manager.get_collection("compounds")
        flask_collection = self.db_manager.get_collection("flasks")
        structure_collection = self.db_manager.get_collection("structures")

        # Reset storage
        self.compounds = {}

        # draw selected compound in center
        o_id = db.ID(self.selected_compound['_id']["$oid"])
        if self.selected_compound["_objecttype"] == "compound":
            self.centroid_compound: Union[db.Compound, db.Flask] = db.Compound(o_id, compound_collection)
        elif self.selected_compound["_objecttype"] == "flask":
            self.centroid_compound = db.Flask(o_id, flask_collection)

        self.centroid_item = Compound(
            0, 0, pen=self.compound_pen, brush=self.compound_brush
        )
        self.centroid_item.db_representation = self.centroid_compound

        self.__bind_functions_to_object(self.centroid_item)
        cid_string = self.centroid_compound.id().string()
        self.compounds[cid_string] = self.centroid_item
        self.scene_object.addItem(self.compounds[cid_string])
        self.item_list[0].append(self.centroid_item)

        # calculate angle depending on number of nodes
        if not self.__unique_structures:
            return
        number_of_structures = len(self.__unique_structures)
        angle_alpha = 360 / number_of_structures
        list_of_angles = []
        for i in range(number_of_structures):
            list_of_angles.append(i * angle_alpha)

        # position of nodes in circle around center
        pos = []
        for angle in list_of_angles:
            pos.append((np.cos(np.pi * (angle / 180)), np.sin(np.pi * (angle / 180))))

        # make circle around center-compound out of structure-nodes
        for i, sid in enumerate(self.__unique_structures):
            s = db.Structure(db.ID(sid))
            s.link(structure_collection)

            x = 250 * pos[i][0]
            y = 250 * pos[i][1]
            self.compounds[i] = Structure(
                x, y, pen=self.structure_pen, brush=self.structure_brush
            )
            self.compounds[i].db_representation = s
            self.__bind_functions_to_object(self.compounds[i])
            self.scene_object.addItem(self.compounds[i])
            self.item_list[1].append(self.compounds[i])

            # draw edges between center node and structures
            side_point = QPoint(self.centroid_item.center())
            path = QPainterPath(self.compounds[i].center())
            path.lineTo(side_point)
            self.scene_object.addPath(path, pen=self.path_pen)

        # Move nodes above lines
        self.__move_to_foreground(self.compounds)
        self.timer.start()

    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            zoomInFactor = 1.25
            zoomOutFactor = 1 / zoomInFactor

            # Save the scene pos
            oldPos = self.mapToScene(event.pos())

            # Zoom
            if event.angleDelta().y() > 0:
                zoomFactor = zoomInFactor
            else:
                zoomFactor = zoomOutFactor
            self.scale(zoomFactor, zoomFactor)

            # Get the new position
            newPos = self.mapToScene(event.pos())

            # Move scene to old position
            delta = newPos - oldPos
            self.translate(delta.x(), delta.y())
        elif event.modifiers() == Qt.ShiftModifier:
            self.horizontalScrollBar().wheelEvent(event)
        else:
            self.verticalScrollBar().wheelEvent(event)

    def __move_to_foreground(self, dictionary: Dict[Union[str, int], QGraphicsItem]) -> None:
        for k in dictionary:
            self.scene_object.removeItem(dictionary[k])
            self.scene_object.addItem(dictionary[k])

    def __bind_functions_to_object(self, object: Any) -> None:
        object.bind_mouse_press_function(self.mouse_press_function)
        object.setAcceptHoverEvents(True)
        object.bind_hover_enter_function(self.hover_enter_function)
        object.bind_hover_leave_function(self.hover_leave_function)
        object.bind_menu_function(self.menu_function)

    def menu_function(self, event, item: QGraphicsItem) -> None:
        menu = QMenu()
        copyid_action = menu.addAction('Copy ID')
        copy_masm_graph_action = menu.addAction('Copy Molassembler Graph')
        if isinstance(item, Structure):
            move_action = menu.addAction('Move to Molecular Viewer')
        _ = menu.addAction('Close Menu')
        action = menu.exec_(event.screenPos())  # type: ignore

        if action == copyid_action:
            current_id = item.db_representation.id().string()
            copy_text_to_clipboard(current_id)
        elif action == copy_masm_graph_action:
            if isinstance(item, Compound):
                s = db.Structure(item.db_representation.get_centroid())
                structures = self.db_manager.get_collection("structures")
                s.link(structures)
            if isinstance(item, Structure):
                s = item.db_representation
            if not s.has_graph("masm_cbor_graph"):
                return
            masm_graph = s.get_graph("masm_cbor_graph")
            copy_text_to_clipboard(masm_graph)
        elif isinstance(item, Structure):
            if action == move_action:
                tab = get_core_tab('molecule_viewer')
                if tab is not None:
                    s = item.db_representation
                    tab.update_molecule(atoms=s.get_atoms())

    def keyPressEvent(self, event):
        if (event.type() == QEvent.KeyPress and event == QKeySequence.Copy):
            if self.__current_focus is not None:
                copy_text_to_clipboard(self.__current_focus)

    def mousePressEvent(self, ev):
        if ev.button() == Qt.LeftButton:
            if not self.itemAt(self.mapToScene(ev.pos()).toPoint()) and self.__current_focus:
                self.__current_focus = None
                self.parent().update_selection_dependent_widgets(None)
                self.__reset_item_colors()
        super().mousePressEvent(ev)

    def __reset_item_colors(self):
        for i in self.item_list[0]:
            i.setPen(self.structure_pen)  # type: ignore
            i.reset_brush()
        for i in self.item_list[1]:
            i.setPen(self.compound_pen)  # type: ignore
            i.reset_brush()

    def mouse_press_function(self, _, item: QGraphicsItem) -> None:
        self.__current_focus = item.db_representation.id().string()
        self.__reset_item_colors()
        item.setPen(self.hover_pen)
        item.setBrush(self.hover_brush)
        item.update()
        if isinstance(item, Compound):
            self.parent().update_selection_dependent_widgets(None)
        elif isinstance(item, Structure):
            self.parent().update_selection_dependent_widgets(self.__current_focus)

    def hover_enter_function(self, _, item: QGraphicsItem) -> None:
        item_db_id_h = item.db_representation.id().string()
        if isinstance(item, Structure):
            self.parent().update_plot_only(
                self.__current_focus,
                additional_highlights=[item_db_id_h]
            )
        if item_db_id_h != self.__current_focus:
            item.setPen(self.hover_pen)
            item.setBrush(self.hover_brush)
            item.update()

    def hover_leave_function(self, _, item: QGraphicsItem) -> None:
        item_db_id = item.db_representation.id().string()
        self.parent().update_plot_only(self.__current_focus, additional_highlights=[])
        if item_db_id != self.__current_focus:
            if isinstance(item, Compound):
                item.setPen(self.compound_pen)
                item.setBrush(self.compound_brush)
            elif isinstance(item, Structure):
                item.setPen(self.structure_pen)
                item.setBrush(self.structure_brush)
            item.update()


class ExpandReaction(QWidget):

    def __init__(self, parent: Optional[QWidget], db_manager: db.Manager,
                 selected_aggregate: Dict[str, Any]) -> None:
        super(ExpandReaction, self).__init__(parent=parent)

        self.__selected_aggregate = selected_aggregate
        self.__structures = db_manager.get_collection('structures')
        self.__elementary_steps = db_manager.get_collection('elementary_steps')

        self.__molecular_viewer = MoleculeWidget(
            parent=self,
            alternative_zoom_controls=True,
            disable_modification=True
        )
        self.__molecular_viewer.setMinimumHeight(200)
        self.__molecular_viewer.setMinimumWidth(200)
        self.__molecule_widget = self.__molecular_viewer
        self.__figure = Figure(figsize=(5, 5))
        self.__plot_widget = FigureCanvasQTAgg(self.__figure)
        self.__plot_widget.setMinimumHeight(200)
        self.__plot_widget.setMinimumWidth(200)
        self.__network_widget = ExpandedSingleReactionNetwork(self, db_manager, selected_aggregate)
        self.__network_widget.setMinimumHeight(400)
        self.__network_widget.setMinimumWidth(400)

        self.__vbox_widget = QWidget()
        self.__vbox = QVBoxLayout()
        self.__vbox.addWidget(self.__molecule_widget)
        self.__vbox.addWidget(self.__plot_widget)
        self.__vbox_widget.setLayout(self.__vbox)
        self.__layout = QHBoxLayout()
        self.__layout.addWidget(self.__network_widget)
        self.__layout.addWidget(self.__vbox_widget)
        self.setLayout(self.__layout)

        self.__elementary_step_ids: List[str] = []
        self.__compound_ids: List[str] = []
        self.__mep_splines_fits_by_id: Dict[str, Optional[Any]] = {}
        self.__min_energy: float = 0.0
        self.__number_of_highlights: int = 0
        self.__spline_order: int = 3
        self.__prev_point = None
        self.__current_focus: Optional[str] = None
        self.__gather_data()
        self.__init_plot()

    def __gather_data(self) -> None:
        for cid in [x['id']['$oid'] for x in self.__selected_aggregate['lhs']]:
            self.__compound_ids.append(cid)
        for cid in [x['id']['$oid'] for x in self.__selected_aggregate['rhs']]:
            self.__compound_ids.append(cid)
        lhs_energies: List[float] = []
        for sid in [x['$oid'] for x in self.__selected_aggregate['elementary_steps']]:
            step = db.ElementaryStep(db.ID(sid))
            step.link(self.__elementary_steps)
            self.__elementary_step_ids.append(sid)
            if step.has_spline():
                spline = step.get_spline()
                lhs_energies.append(spline.evaluate(0.0, self.__spline_order)[0])
            else:
                spline = None
            self.__mep_splines_fits_by_id[sid] = spline
        if lhs_energies:
            self.__min_energy = min(lhs_energies)

    def __init_plot(self):
        self.__ax = self.__figure.add_subplot(1, 1, 1)
        color_figure(self.__figure)
        font = get_font()
        color_axis(self.__ax)
        self.__ax.set_title("Minimum Energy Paths", font)
        self.__ax.set_ylabel("Relative Electronic Energy in kJ/mol", font)
        self.__ax.set_xlabel(
            f"Relative to: {self.__min_energy:.5f} Hartree"
        )
        x = [i / 1000 for i in range(1001)]
        for spline in self.__mep_splines_fits_by_id.values():
            if spline is None:
                continue
            y = [(spline.evaluate(i / 1000, self.__spline_order)[0] - self.__min_energy) * 2625.5 for i in range(1001)]
            self.__ax.plot(x, y, color=get_primary_line_color())
        self.__plot_widget.draw()

    def __update_plot_highlights(self, highlights: Optional[List[str]] = None):
        x = [i / 1000 for i in range(1001)]
        for _ in range(self.__number_of_highlights):
            self.__ax.lines[-1].remove()
        self.__number_of_highlights = 0
        if highlights is None:
            return
        for sid in highlights:
            if sid not in self.__mep_splines_fits_by_id:
                continue
            spline = self.__mep_splines_fits_by_id[sid]
            if spline is None:
                continue
            y = [(spline.evaluate(i / 1000, self.__spline_order)[0] - self.__min_energy) * 2625.5 for i in range(1001)]
            self.__ax.plot(x, y, color="#D40000", linewidth=2)
            self.__number_of_highlights += 1
        self.__plot_widget.draw()
        if self.__current_focus is None:
            self.__draw_point(0)

    def __draw_point(self, current_frame: int):
        if self.__prev_point is not None:
            self.__prev_point.remove()
            self.__prev_point = None
        if self.__current_focus is None:
            return
        if self.__current_focus not in self.__elementary_step_ids:
            return
        x = current_frame / 49.0
        spline = self.__mep_splines_fits_by_id[self.__current_focus]
        if spline is not None:
            spline_value = spline.evaluate(x, self.__spline_order)[0]
            y = (spline_value - self.__min_energy) * 2625.5
            self.__prev_point = self.__ax.scatter(x, y, color="C1", zorder=100)
            self.__plot_widget.draw()

    @staticmethod
    def __spline_to_trajectory(spline: utils.bsplines.TrajectorySpline, n_frames: int = 50) \
            -> utils.MolecularTrajectory:
        trajectory = utils.MolecularTrajectory()
        _, atoms = spline.evaluate(0, 3)
        trajectory.elements = atoms.elements
        for i in range(n_frames):
            p = i / n_frames
            energy, atoms = spline.evaluate(p, 3)
            trajectory.push_back(atoms.positions, energy)
        return trajectory

    def __update_molecule_widget(self, current_focus: Optional[str]) -> None:
        if current_focus is not None and current_focus in self.__elementary_step_ids:
            trajectory: Any = utils.MolecularTrajectory()
            spline = self.__mep_splines_fits_by_id[current_focus]
            if spline is not None:
                trajectory = self.__spline_to_trajectory(spline)
            else:
                # barrierless reaction
                s = utils.AtomCollection()
                trajectory.elements = s.elements
                trajectory.push_back(s.positions)
            old_widget = self.__molecule_widget
            n = self.__vbox.indexOf(self.__molecule_widget)
            self.__molecule_widget = MoleculeVideo(
                parent=self,
                trajectory=trajectory,
                mol_widget=self.__molecular_viewer
            )
            if self.__molecule_widget.slider is not None:
                self.__molecule_widget.slider.valueChanged.connect(self.__draw_point)  # pylint: disable=no-member
            self.__vbox.insertWidget(n, self.__molecule_widget)
            if isinstance(old_widget, MoleculeVideo):
                old_widget.close()
            self.__draw_point(0)
            return
        elif current_focus in self.__compound_ids or current_focus is None:
            atoms = utils.AtomCollection()
        else:
            structure = db.Structure(db.ID(current_focus))
            structure.link(self.__structures)
            atoms = structure.get_atoms()
        old_widget = self.__molecule_widget
        n = self.__vbox.indexOf(self.__molecule_widget)
        self.__molecule_widget = self.__molecular_viewer
        self.__vbox.insertWidget(n, self.__molecule_widget)
        self.__molecular_viewer.update_molecule(atoms=atoms)
        if isinstance(old_widget, MoleculeVideo):
            old_widget.close()

    def update_selection_dependent_widgets(
        self,
        current_focus: Optional[str] = None,
        current_hover: Optional[str] = None
    ):
        highlights = []
        if current_focus is not None:
            highlights.append(current_focus)
        if current_hover is not None and current_hover != current_focus:
            highlights.append(current_hover)
        if self.__current_focus != current_focus:
            self.__current_focus = current_focus
            self.__update_molecule_widget(current_focus)
        self.__update_plot_highlights(highlights)


class ExpandedSingleReactionNetwork(QGraphicsView):
    # window regarding expand compound
    def __init__(self, parent: Optional[QWidget], db_manager: db.Manager, selected_reaction: dict) -> None:
        self.window_width = 1296   # golden ratio: 800 * 1,62 = 1296
        self.window_height = 800
        self.scene_object = QGraphicsScene(0, 0, self.window_width, self.window_height)
        super(ExpandedSingleReactionNetwork, self).__init__(self.scene_object, parent=parent)

        self.setMouseTracking(True)
        self.setInteractive(True)

        # settings regarding presentation
        self.compound_color = qcolor_by_key('compoundColor')
        self.highlight_color = qcolor_by_key('highlightColor')
        self.border_color = qcolor_by_key('borderColor')
        self.edge_color = qcolor_by_key('edgeColor')
        self.structure_color = qcolor_by_key('structureColor')
        self.elementary_step_color = qcolor_by_key('elementaryStepColor')

        self.structure_brush = build_brush(self.structure_color)
        self.structure_pen = build_pen(self.border_color)
        self.compound_brush = build_brush(self.compound_color)
        self.compound_pen = build_pen(self.border_color)
        self.elementary_step_brush = build_brush(self.elementary_step_color)
        self.elementary_step_pen = build_pen(self.border_color)
        self.hover_brush = build_brush(self.highlight_color)
        self.hover_pen = build_pen(self.highlight_color, width=2)
        self.path_pen = build_pen(self.edge_color)

        # rendering smoother lines and edges of nodes
        self.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)

        # Add all data
        self.__current_focus: Optional[str] = None
        self.__current_hover: Optional[str] = None
        self.recolored_items: List[List[db.ID]] = [[], [], [], []]
        # TODO: this is broken
        self.selected_reaction = selected_reaction
        self.total_number_compound_rhs = self.selected_reaction["rhs"]
        self.total_number_compound_lhs = self.selected_reaction["lhs"]
        self.total_number_elementary_steps = self.selected_reaction["elementary_steps"]

        self.total_number_structures_rhs: List[dict] = []
        self.total_number_structures_lhs: List[dict] = []

        self.line_items1: Dict[str, Any] = {}
        self.line_items2: Dict[str, Any] = {}
        self.line_items4: Dict[str, Any] = {}
        self.line_items5: Dict[str, Any] = {}
        self.line_lists = [self.line_items1, self.line_items5, self.line_items2, self.line_items4]

        self.db_manager: db.Manager = db_manager
        self.compounds_rhs: Dict[str, Compound] = {}
        self.compounds_lhs: Dict[str, Compound] = {}
        self.structure_rhs: Dict[str, Structure] = {}
        self.structure_lhs: Dict[str, Structure] = {}
        self.elementary: Dict[str, Reaction] = {}

        self.plot_network_reaction()

    def __getitem__(self, i):
        # get information of database
        # to make subscriptable
        return f"{i}"

    def plot_network_reaction(self) -> None:
        # Create singular timer for all added animated items
        self.timer = QTimeLine(1000)
        self.frames = 40
        self.timer.setFrameRange(0, self.frames)

        # db-info of selected compound
        structure_collection = self.db_manager.get_collection("structures")
        compound_collection = self.db_manager.get_collection("compounds")
        flask_collection = self.db_manager.get_collection("flasks")
        elementary_step_collection = self.db_manager.get_collection("elementary_steps")

        # Reset storage
        self.lhs_node_types = []
        self.rhs_node_types = []
        self.compounds_rhs = {}
        self.structure_rhs = {}
        self.elementary = {}
        self.compounds_lhs = {}
        self.structure_lhs = {}

        # Layers #
        pos_elementary_steps_x = self.window_width / 2
        pos_compounds_rhs_x = 40
        pos_structures_rhs_x = (self.window_width / 2 + pos_compounds_rhs_x) / 2
        pos_structures_lhs_x = (3 * self.window_width / 2 - pos_compounds_rhs_x) / 2
        pos_compounds_lhs_x = self.window_width - pos_compounds_rhs_x

        # structure rhs
        list_add_rhs = []
        for rhs_element in self.selected_reaction["rhs"]:
            list_add_rhs.append(rhs_element["id"])
            self.rhs_node_types.append(rhs_element["type"])
            rhs_id = db.ID(rhs_element["id"]["$oid"])
            if rhs_element["type"] == "compound":
                ag: Union[db.Compound, db.Flask] = db.Compound(rhs_id, compound_collection)
                rhs_aggregate = json.loads(str(ag))
            elif rhs_element["type"] == "flask":
                ag = db.Flask(rhs_id, flask_collection)
                rhs_aggregate = json.loads(str(ag))
            else:
                raise RuntimeError("Unknown aggregate type in database.")
            self.total_number_structures_rhs.extend(rhs_aggregate["structures"])
            list_add_rhs.append(rhs_aggregate["structures"])

        self.total_number_structures_rhs = [dict(tupleized) for tupleized in set(
            tuple(item.items()) for item in self.total_number_structures_rhs)]  # reduce duplicated structure

        pos3_y = self.window_height / 2
        all_structure_ids = set()
        for i in range(len(self.total_number_elementary_steps)):
            if i % 2 == 0 or i == 0:
                pos3_y = pos3_y - i * 30
            else:
                pos3_y = pos3_y + i * 30

            elementary_reaction_step = db.ElementaryStep(db.ID(self.total_number_elementary_steps[i]["$oid"]),
                                                         elementary_step_collection)
            self.elementary_reaction_step_item = Reaction(
                pos_elementary_steps_x, pos3_y, pen=self.elementary_step_pen, brush=self.elementary_step_brush
            )
            self.elementary_reaction_step_item.db_representation = elementary_reaction_step
            lhs, rhs = elementary_reaction_step.get_reactants(db.Side.BOTH)
            for lhs_id in lhs:
                all_structure_ids.add(str(lhs_id))
            for rid in rhs:
                all_structure_ids.add(str(rid))

            self.__bind_functions_to_object(self.elementary_reaction_step_item)
            elemen_string = elementary_reaction_step.id().string()
            self.elementary[elemen_string] = self.elementary_reaction_step_item
            self.scene_object.addItem(self.elementary[elemen_string])

        # layer 1 for compound_rhs
        for i in range(len(self.total_number_compound_rhs)):
            pos1_y = self.window_height / 2
            if len(self.total_number_compound_rhs) == 1:
                pos1_y = self.window_height / 2
            else:
                if i % 2 == 0 or i == 0:
                    pos1_y -= 150
                else:
                    pos1_y += 150
            type_str = self.total_number_compound_rhs[i]["type"]
            a_id = db.ID(self.total_number_compound_rhs[i]["id"]["$oid"])
            a_type = db.CompoundOrFlask.COMPOUND if type_str == "compound" else db.CompoundOrFlask.FLASK
            compound_in = get_compound_or_flask(a_id, a_type, compound_collection, flask_collection)
            self.compound_in_item = Compound(
                pos_compounds_rhs_x, pos1_y, pen=self.compound_pen, brush=self.compound_brush
            )
            self.compound_in_item.db_representation = compound_in

            self.__bind_functions_to_object(self.compound_in_item)
            cid_string = compound_in.id().string()
            self.compounds_rhs[cid_string] = self.compound_in_item
            self.scene_object.addItem(self.compounds_rhs[cid_string])

        # layer 2 for structure_rhs
        pos2_y = self.window_height / 2
        relevant_structures_rhs = []
        for i in range(len(self.total_number_structures_rhs)):
            if self.total_number_structures_rhs[i]['$oid'] in all_structure_ids:
                relevant_structures_rhs.append(self.total_number_structures_rhs[i])

        for i, relevant_structure in enumerate(relevant_structures_rhs):
            if i % 2 == 0 or i == 0:
                pos2_y -= i * 50
            else:
                pos2_y += i * 50

            structure_in = db.Structure(db.ID(relevant_structure['$oid']), structure_collection)
            structure_in_item = Structure(
                pos_structures_rhs_x, pos2_y, pen=self.structure_pen, brush=self.structure_brush
            )
            structure_in_item.db_representation = structure_in

            self.__bind_functions_to_object(structure_in_item)
            rhs_string = structure_in.id().string()
            self.structure_rhs[rhs_string] = structure_in_item
            self.scene_object.addItem(self.structure_rhs[rhs_string])

            # add lines layer 12
            ids = [value["id"]['$oid'] for value in self.total_number_compound_rhs]
            side_point = QPoint(structure_in_item.center())
            rid_str = structure_in_item.db_representation.id().string()
            for k, c in enumerate(ids):
                if c == list_add_rhs[0]['$oid']:
                    for j in range(len(list_add_rhs[1])):
                        if rid_str in list_add_rhs[1][j]['$oid']:
                            path = QPainterPath(self.compounds_rhs[c].center())
                            path.lineTo(side_point)
                            lid = rid_str + "_" + c + "_" + str(k)
                            self.line_items1[lid] = self.scene_object.addPath(path, pen=self.path_pen)

                elif c == list_add_rhs[2]['$oid']:
                    for j in range(len(list_add_rhs[3])):
                        if rid_str in list_add_rhs[3][j]['$oid']:
                            path = QPainterPath(self.compounds_rhs[c].center())
                            path.lineTo(side_point)
                            lid = rid_str + "_" + c + "_" + str(k)
                            self.line_items1[lid] = self.scene_object.addPath(path, pen=self.path_pen)

        # layer 5 for compound_lhs
        self.list_compound_out_item = []
        for i in range(len(self.total_number_compound_lhs)):
            pos5_y = self.window_height / 2
            if i % 2 == 0 or i == 0:
                pos5_y += 150
            else:
                pos5_y -= 150

            a_id = db.ID(self.total_number_compound_lhs[i]["id"]["$oid"])
            type_str = self.total_number_compound_lhs[i]["type"]
            a_type = db.CompoundOrFlask.COMPOUND if type_str == "compound" else db.CompoundOrFlask.FLASK
            compound_out = get_compound_or_flask(a_id, a_type, compound_collection, flask_collection)
            compound_out_item = Compound(
                pos_compounds_lhs_x, pos5_y, pen=self.compound_pen, brush=self.compound_brush
            )

            # list of compounds for layer 45 in next loop
            self.list_compound_out_item.append(compound_out_item)

            compound_out_item.db_representation = compound_out

            self.__bind_functions_to_object(compound_out_item)
            comin_string = compound_out.id().string()
            self.compounds_lhs[comin_string] = compound_out_item
            self.scene_object.addItem(self.compounds_lhs[comin_string])

        # structure lhs
        list_add_lhs = []
        for lhs_element in self.selected_reaction["lhs"]:
            list_add_lhs.append(lhs_element["id"])
            self.lhs_node_types.append(lhs_element["type"])

            lhs_id = db.ID(lhs_element["id"]["$oid"])
            if lhs_element["type"] == "compound":
                ag = db.Compound(lhs_id, compound_collection)
                lhs_aggregate = json.loads(str(ag))
            elif lhs_element["type"] == "flask":
                ag = db.Flask(lhs_id, flask_collection)
                lhs_aggregate = json.loads(str(ag))
            else:
                raise RuntimeError("Unknown aggregate type in database.")
            self.total_number_structures_lhs.extend(lhs_aggregate["structures"])
            list_add_lhs.append(lhs_aggregate["structures"])

        self.total_number_structures_lhs = [dict(tupleized) for tupleized in set(
            tuple(item.items()) for item in self.total_number_structures_lhs)]  # reduce duplicated structure

        # layer 4 for structure_lhs
        pos4_y = self.window_height / 2
        relevant_structures_lhs = []
        for i in range(len(self.total_number_structures_lhs)):
            if self.total_number_structures_lhs[i]['$oid'] in all_structure_ids:
                relevant_structures_lhs.append(self.total_number_structures_lhs[i])

        for i, relevant_structure in enumerate(relevant_structures_lhs):
            if i % 2 == 0 or i == 0:
                pos4_y -= i * 50
            else:
                pos4_y += i * 50

            structure_out = db.Structure(db.ID(relevant_structure['$oid']), structure_collection)
            self.structure_out_item = Structure(
                pos_structures_lhs_x, pos4_y, pen=self.structure_pen, brush=self.structure_brush
            )
            self.structure_out_item.db_representation = structure_out

            self.__bind_functions_to_object(self.structure_out_item)
            lhs_string = structure_out.id().string()
            self.structure_lhs[lhs_string] = self.structure_out_item
            self.scene_object.addItem(self.structure_lhs[lhs_string])

            # add lines layer 45
            ids = [value["id"]['$oid'] for value in self.total_number_compound_lhs]
            side_point = QPoint(self.structure_out_item.center())
            rid_str = self.structure_out_item.db_representation.id().string()
            for k, c in enumerate(ids):
                if c == list_add_lhs[0]['$oid']:
                    for j in range(len(list_add_lhs[1])):
                        if rid_str in list_add_lhs[1][j]['$oid']:
                            path = QPainterPath(self.compounds_lhs[c].center())
                            path.lineTo(side_point)
                            lid = rid_str + "_" + c + "_" + str(k)
                            self.line_items5[lid] = self.scene_object.addPath(path, pen=self.path_pen)

                elif c == list_add_lhs[2]['$oid']:
                    for j in range(len(list_add_lhs[3])):
                        if rid_str in list_add_lhs[3][j]['$oid']:
                            path = QPainterPath(self.compounds_lhs[c].center())
                            path.lineTo(side_point)
                            lid = rid_str + "_" + c + "_" + str(k)
                            self.line_items5[lid] = self.scene_object.addPath(path, pen=self.path_pen)

        pos3_y = self.window_height / 2
        for i in range(len(self.total_number_elementary_steps)):
            if i % 2 == 0 or i == 0:
                pos3_y = pos3_y - i * 30
            else:
                pos3_y = pos3_y + i * 30

            elementary_reaction_step = db.ElementaryStep(db.ID(self.total_number_elementary_steps[i]["$oid"]),
                                                         elementary_step_collection)
            self.elementary_reaction_step_item = Reaction(
                pos_elementary_steps_x, pos3_y, pen=self.elementary_step_pen, brush=self.elementary_step_brush
            )
            self.elementary_reaction_step_item.db_representation = elementary_reaction_step
            reactant_structures = elementary_reaction_step.get_reactants(db.Side.BOTH)
            self.elementary_reaction_step_item.lhs_ids = [db_id.string() for db_id in reactant_structures[0]]
            self.elementary_reaction_step_item.rhs_ids = [db_id.string() for db_id in reactant_structures[1]]

            self.__bind_functions_to_object(self.elementary_reaction_step_item)
            elemen_string = elementary_reaction_step.id().string()
            self.elementary[elemen_string] = self.elementary_reaction_step_item
            self.elementary[elemen_string].setAcceptHoverEvents(True)
            self.elementary[elemen_string].bind_hover_enter_function(
                self.hover_enter_function
            )
            self.elementary[elemen_string].bind_hover_leave_function(
                self.hover_leave_function
            )
            self.scene_object.addItem(self.elementary[elemen_string])

        for i, key in enumerate(self.elementary):
            elementary_step_item = self.elementary[key]
            # add lines layer 34
            for lhs_id_str in elementary_step_item.lhs_ids:
                if lhs_id_str in self.structure_lhs:
                    side_point = QPoint(elementary_step_item.outgoing())
                    path = QPainterPath(self.structure_lhs[lhs_id_str].center())
                    path.lineTo(side_point)
                    lid = key + "_" + lhs_id_str + "_" + str(i)
                    self.line_items4[lid] = self.scene_object.addPath(path, pen=self.path_pen)

            # add lines layer 23
            for rhs_id_str in elementary_step_item.rhs_ids:
                if rhs_id_str in self.structure_rhs:
                    side_point = QPoint(elementary_step_item.incoming())
                    path = QPainterPath(self.structure_rhs[rhs_id_str].center())
                    path.lineTo(side_point)
                    lid = key + "_" + rhs_id_str + "_" + str(i)
                    self.line_items2[lid] = self.scene_object.addPath(path, pen=self.path_pen)

        # move nodes above lines
        self.__move_to_foreground(self.compounds_rhs)
        self.__move_to_foreground(self.structure_rhs)
        self.__move_to_foreground(self.elementary)
        self.__move_to_foreground(self.structure_lhs)
        self.__move_to_foreground(self.compounds_lhs)

        self.timer.start()

        self.scene_object.update()
        self.view = QGraphicsView(self.scene_object)
        # rendering smoother lines and edges of nodes
        self.view.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)

    def __move_to_foreground(self, dictionary: Dict[str, QGraphicsItem]) -> None:
        for k in dictionary:
            self.scene_object.removeItem(dictionary[k])
            self.scene_object.addItem(dictionary[k])

    def __bind_functions_to_object(self, object: Any) -> None:
        object.bind_mouse_press_function(self.mouse_press_function)
        object.setAcceptHoverEvents(True)
        object.bind_hover_enter_function(self.hover_enter_function)
        object.bind_hover_leave_function(self.hover_leave_function)
        object.bind_menu_function(self.menu_function)

    def menu_function(self, event, item: QGraphicsItem) -> None:
        menu = QMenu()
        copyid_action = menu.addAction('Copy ID')
        if isinstance(item, Structure) or isinstance(item, Compound):
            copy_masm_graph_action = menu.addAction('Copy Molassembler Graph')
        if isinstance(item, Structure):
            move_action = menu.addAction('Move to Molecular Viewer')
        if isinstance(item, Reaction):
            if item.db_representation.has_transition_state():
                move_ts_action = menu.addAction('Move TS to Molecular Viewer')
        _ = menu.addAction('Close Menu')
        action = menu.exec_(event.screenPos())  # type: ignore

        if action == copyid_action:
            current_id = item.db_representation.id().string()
            copy_text_to_clipboard(current_id)
        if isinstance(item, Structure) or isinstance(item, Compound):
            if action == copy_masm_graph_action:
                if isinstance(item, Compound):
                    s = db.Structure(item.db_representation.get_centroid())
                    structures = self.db_manager.get_collection("structures")
                    s.link(structures)
                if isinstance(item, Structure):
                    s = item.db_representation
                if not s.has_graph("masm_cbor_graph"):
                    return
                masm_graph = s.get_graph("masm_cbor_graph")
                copy_text_to_clipboard(masm_graph)
        if isinstance(item, Structure):
            if action == move_action:
                tab = get_core_tab('molecule_viewer')
                if tab is not None:
                    s = item.db_representation
                    tab.update_molecule(atoms=s.get_atoms())
        if isinstance(item, Reaction):
            if item.db_representation.has_transition_state():
                if action == move_ts_action:
                    tab = get_core_tab('molecule_viewer')
                    if tab is not None:
                        es = item.db_representation
                        tsid = es.get_transition_state()
                        s = db.Structure(tsid)
                        structures = self.db_manager.get_collection("structures")
                        s.link(structures)
                        tab.update_molecule(atoms=s.get_atoms())

    def __reset_item_colors(self):
        for i in self.recolored_items[0]:
            i.setPen(self.structure_pen)  # type: ignore
            i.reset_brush()
        for i in self.recolored_items[1]:
            i.setPen(self.compound_pen)  # type: ignore
            i.reset_brush()
        for i in self.recolored_items[2]:
            i.setPen(self.elementary_step_pen)  # type: ignore
            i.reset_brush()
        for i in self.recolored_items[3]:
            i.setPen(self.path_pen)  # type: ignore
        self.recolored_items = [[] for _ in self.recolored_items]

    def mouse_press_function(self, _, item: QGraphicsItem) -> None:
        if self.__current_focus == item.db_representation.id().string():
            return
        self.__current_focus = item.db_representation.id().string()
        self.parent().update_selection_dependent_widgets(self.__current_focus, self.__current_hover)
        self.__reset_item_colors()
        item.setPen(self.hover_pen)
        item.setBrush(self.hover_brush)
        item.update()
        if isinstance(item, Compound):
            self.recolored_items[1].append(item)
        elif isinstance(item, Reaction):
            self.recolored_items[2].append(item)
        elif isinstance(item, Structure):
            self.recolored_items[0].append(item)
        for line_list in self.line_lists:
            for key in line_list.keys():
                if self.__current_focus in key:  # type: ignore
                    line_list[key].setPen(self.hover_pen)  # type: ignore
                    self.recolored_items[3].append(line_list[key])

    def hover_enter_function(self, _, item: QGraphicsItem) -> None:
        self.__current_hover = item.db_representation.id().string()
        self.parent().update_selection_dependent_widgets(self.__current_focus, self.__current_hover)
        item_db_id_h = item.db_representation.id().string()

        if item_db_id_h != self.__current_focus:
            item.setPen(self.hover_pen)
            item.setBrush(self.hover_brush)
            item.update()
            if isinstance(item, Compound):
                for j in range(len(self.line_lists[:2])):
                    for k in self.line_lists[:2][j].keys():
                        if item_db_id_h in k:
                            self.line_lists[:2][j][k].setPen(self.hover_pen)
            elif isinstance(item, Reaction):
                # Key of selected elementary_reaction_step
                key_item = [key for key, value in self.elementary.items() if value == item]
                for j in range(len(self.line_lists[2:])):
                    for k in self.line_lists[2:][j].keys():
                        if (key_item[0]) in k:
                            self.line_lists[2:][j][k].setPen(self.hover_pen)
            elif isinstance(item, Structure):
                for j in range(len(self.line_lists)):
                    for k in self.line_lists[j].keys():
                        if item_db_id_h in k:
                            self.line_lists[j][k].setPen(self.hover_pen)

    def hover_leave_function(self, _, item: QGraphicsItem) -> None:
        self.__current_hover = None
        self.parent().update_selection_dependent_widgets(self.__current_focus, self.__current_hover)
        item_db_id = item.db_representation.id().string()

        if item_db_id != self.__current_focus:
            if isinstance(item, Compound):
                item.setPen(self.compound_pen)
                item.setBrush(self.compound_brush)
            elif isinstance(item, Reaction):
                item.setPen(self.elementary_step_pen)
                item.setBrush(self.elementary_step_brush)
            elif isinstance(item, Structure):
                item.setPen(self.structure_pen)
                item.setBrush(self.structure_brush)

            for j in range(len(self.line_lists)):
                for k in self.line_lists[j].keys():
                    if item_db_id in k:
                        self.line_lists[j][k].setPen(self.path_pen)
            item.update()

    def keyPressEvent(self, event):
        if (event.type() == QEvent.KeyPress and event == QKeySequence.Copy):
            if self.__current_focus is not None:
                copy_text_to_clipboard(self.__current_focus)

    def mousePressEvent(self, ev):
        if ev.button() == Qt.LeftButton:
            if not self.itemAt(self.mapToScene(ev.pos()).toPoint()) and self.__current_focus:
                self.__current_focus = None
                self.parent().update_selection_dependent_widgets(self.__current_focus, self.__current_hover)
                self.__reset_item_colors()
        super().mousePressEvent(ev)

    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            zoomInFactor = 1.25
            zoomOutFactor = 1 / zoomInFactor

            # Save the scene pos
            oldPos = self.mapToScene(event.pos())

            # Zoom
            if event.angleDelta().y() > 0:
                zoomFactor = zoomInFactor
            else:
                zoomFactor = zoomOutFactor
            self.scale(zoomFactor, zoomFactor)

            # Get the new position
            newPos = self.mapToScene(event.pos())

            # Move scene to old position
            delta = newPos - oldPos
            self.translate(delta.x(), delta.y())
        elif event.modifiers() == Qt.ShiftModifier:
            self.horizontalScrollBar().wheelEvent(event)
        else:
            self.verticalScrollBar().wheelEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import List, Set


class OrbitalGroup:
    def __init__(self, n_systems: int):
        if n_systems == 0:
            raise RuntimeError("There must be at least one system in an orbital group.")
        self._system_wise_indices: List[Set[int]] = [set() for _ in range(n_systems)]

    def n_systems(self):
        return len(self._system_wise_indices)

    def add_orbitals(self, new_indices: List[int]):
        if self.n_systems() != len(new_indices):
            raise RuntimeError("The number of systems is inconsistent in the orbital mapping.")
        for i_sys, i in enumerate(new_indices):
            self._system_wise_indices[i_sys].add(i)

    def get_indices_for_system(self, i_sys: int) -> Set[int]:
        if i_sys >= self.n_systems():
            raise RuntimeError("System index out of bounds.")
        return self._system_wise_indices[i_sys]

    def n_orbitals(self):
        return len(self._system_wise_indices[0])

    def empty(self):
        return self.n_orbitals() == 0


class OrbitalGroupMap:
    def __init__(self, orbital_groups: List[OrbitalGroup]):
        self._groups: List[OrbitalGroup] = orbital_groups

    def add_orbital_group(self, orbital_group: OrbitalGroup):
        self._groups.append(orbital_group)

    def get_n_systems(self):
        if not self._groups:
            return 0
        return self._groups[0].n_systems()

    def get_orbital_groups(self):
        return self._groups

    def get_index_sets_for_system(self, i: int) -> List[Set[int]]:
        assert i < self.get_n_systems()
        index_sets = []
        for group in self._groups:
            index_sets.append(group.get_indices_for_system(i))
        return index_sets



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Optional
from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    QDialog,
    QLabel,
    QMessageBox,
)
from scine_database import Manager

from scine_heron.io.text_box import yes_or_no_question
from scine_heron.utilities import write_error_message, write_info_message


class DatabaseConnectionDialog(QDialog):
    def __init__(self, parent: Optional[QObject] = None, db_manager=None) -> None:
        super(DatabaseConnectionDialog, self).__init__(parent)
        self.setWindowTitle("Connect to Database")

        # Prep manager
        self.db_manager = db_manager
        default_ip = "127.0.0.1"
        default_port = "27017"
        default_db_name = "default"
        if self.db_manager is None:
            self.db_manager = Manager()
        elif self.db_manager.has_credentials():
            # Use current values
            c = self.db_manager.get_credentials()
            default_ip = c.hostname
            default_port = str(c.port)
            default_db_name = c.database_name

        # Create widgets
        self.ip_label = QLabel("Server IP:")
        self.port_label = QLabel("Server Port:")
        self.name_label = QLabel("Database Name:")
        self.ip = QLineEdit(default_ip)
        self.port = QLineEdit(default_port)
        self.port.setMaxLength(5)
        self.name = QLineEdit(default_db_name)
        self.button_connect = QPushButton("Connect")
        self.button_disconnect = QPushButton("Disconnect")
        self.button_wipe = QPushButton("Wipe")
        self.button_close = QPushButton("Close")

        # Create layout and add widgets
        layout = QVBoxLayout()
        layout.setObjectName("Connect to Database")
        layout.addWidget(self.ip_label)
        layout.addWidget(self.ip)
        layout.addWidget(self.port_label)
        layout.addWidget(self.port)
        layout.addWidget(self.name_label)
        layout.addWidget(self.name)
        layout.addWidget(self.button_connect)
        layout.addWidget(self.button_disconnect)
        layout.addWidget(self.button_wipe)
        layout.addWidget(self.button_close)
        # Set dialog layout
        self.setLayout(layout)
        self.button_connect.clicked.connect(self.connect_manager)  # pylint: disable=no-member
        self.button_disconnect.clicked.connect(self.disconnect_manager)  # pylint: disable=no-member
        self.button_wipe.clicked.connect(self.wipe_db)  # pylint: disable=no-member
        if not self.db_manager.is_connected():
            self.button_disconnect.setEnabled(self.db_manager.is_connected())
            self.button_wipe.setEnabled(self.db_manager.is_connected())
        self.button_close.clicked.connect(self.close)  # pylint: disable=no-member

    def connect_manager(self) -> None:
        import scine_database as db

        credentials = db.Credentials(
            self.ip.text(), int(self.port.text()), self.name.text()
        )
        if credentials != self.db_manager.get_credentials():
            self.db_manager.set_credentials(credentials)
        try:
            self.db_manager.connect()
            if not self.db_manager.has_collection("calculations"):
                ans = yes_or_no_question(self, f"The database {self.name.text()} holds no collections. Create it")
                if ans:
                    self.db_manager.init()
                else:
                    self.db_manager.disconnect()
                    write_error_message("Database not connected")
            self.button_disconnect.setEnabled(self.db_manager.is_connected())
            self.button_wipe.setEnabled(self.db_manager.is_connected())
            self.close()
        except Exception as e:  # pylint: disable=broad-except
            error_dialog = QMessageBox(parent=self)
            error_dialog.setWindowTitle("Error: Database Connection")
            c = self.db_manager.get_credentials()
            if c.port > 65535:
                error_dialog.setText("Invalid Port Number")
            else:
                error_dialog.setText(str(e))
            error_dialog.exec()
        except BaseException as e:
            raise RuntimeError("Caught unknown error in SCINE DB wrapper.") from e

    def disconnect_manager(self) -> None:
        self.db_manager.disconnect()
        self.button_disconnect.setEnabled(False)
        self.button_wipe.setEnabled(False)

    def wipe_db(self) -> None:
        answer = yes_or_no_question(self,
                                    f"Are you sure you want to delete everything in the database '{self.name.text()}'")
        if not answer:
            write_info_message(f"Aborted deletion of {self.name.text()}")
            return
        self.db_manager.wipe()
        self.db_manager.init()

    def get_db_manager(self) -> Manager:
        self.exec_()
        return self.db_manager



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from scine_utilities import (
    StructuralCompletion,
    ANGSTROM_PER_BOHR,
    ElementInfo,
)

from vtk import vtkMolecule
import numpy as np
from typing import List, Optional, Callable, Tuple, Dict, Any

from scine_heron.edit_molecule.collision import collision_multiple

import itertools


def _create_writeable_candidates(n: int) -> List[np.ndarray]:
    """
    Creates `n` np arrays, writable, of shape (3,1),
    to use as output for the methods of StructuralCompletion.
    """
    # Each array must be a different object in memory
    pos_candidates = [np.zeros((3, 1)) for _ in range(n)]
    for c in pos_candidates:
        c.setflags(write=True)
    return pos_candidates


def _get_covalent_radius(z: int) -> float:
    return ElementInfo.covalent_radius(ElementInfo.element(z)) * ANGSTROM_PER_BOHR


def _create_molecule_validator(
        new_n: int, molecule: vtkMolecule
) -> Callable[[np.ndarray], bool]:
    """
    Creates a validator for positions
    that can be used to check if the new atom
    (having atomic number "new_n")
    collides with the molecule.
    """
    # In vtk atoms are represented as balls
    # having a radius of .3,
    SAFETY_MARGIN = 0.35

    other_positions = [
        molecule.GetAtom(i).GetPosition()  # Do not reformat
        for i in range(molecule.GetNumberOfAtoms())
    ]
    other_ns = [
        molecule.GetAtom(i).GetAtomicNumber()  # Do not reformat
        for i in range(molecule.GetNumberOfAtoms())
    ]

    visual_radius_new_atom = SAFETY_MARGIN * _get_covalent_radius(new_n)
    visual_radii = [
        SAFETY_MARGIN * _get_covalent_radius(n) for n in other_ns
    ]

    def validator(position: np.ndarray) -> bool:
        assert position.shape == (3,)
        return not collision_multiple(
            radius=visual_radius_new_atom,
            position=position,
            radii=visual_radii,
            positions=other_positions,
        )

    return validator


def _get_pos_array(atom_id: int, molecule: vtkMolecule) -> np.ndarray:
    """
    Gets the position of the "atom_id"-th atom in the molecule
    as a ndarray.
    Needs to be of (3,1) shape to be used with the StructuralCompletion functions.
    """

    pos = molecule.GetAtom(atom_id).GetPosition()
    return np.array(pos).reshape((3, 1))


def generate_random_position_around_atom(*positions: np.ndarray) -> None:
    """
    Produces positions on the faces, edges and corners of a cube of side 2,
    sorted by distance to the center.
    """
    xs = ys = zs = [-1, 0, 1]
    suggestions = sorted(
        itertools.product(xs, ys, zs),  # producing a lattice
        key=lambda v: sum(x * x for x in v),
    )  # sort by distance
    # Skipping the centre
    for pos, suggestion in zip(positions, suggestions[1:]):
        # assigning to output arrays
        pos[:, 0] = suggestion


# Map between
# number of bonds given ->  list of (function to try, number or outputs of said function)
completion_functions_to_try: Dict[int, Any] = {
    0: [(generate_random_position_around_atom, 26)],
    1: [
        (StructuralCompletion.generate_two_triangle_corners_from_one, 2),
        (StructuralCompletion.generate_three_tetrahedron_corners_from_one, 3),
    ],
    2: [
        (StructuralCompletion.generate_one_triangle_corner_from_two, 1),
        (StructuralCompletion.generate_two_tetrahedron_corners_from_two, 2),
    ],
    3: [(StructuralCompletion.generate_one_tetrahedron_corner_from_three, 1)],
}


def new_atom_position_from_id_list(
        atom_ids: List[int], new_atom_n: int, molecule: vtkMolecule
) -> Tuple[Optional[np.ndarray], Optional[str]]:
    """
    Tries all the methods in scine_utilities
    until a possible position is found
    that does not collide with other atoms in the molecule.
    Returns also the name of the method that worked.

    The first element of the `atom_ids` list
    must be the base atom,
    the next ones must be the `ids` of the atoms
    that are bond to it.

    The vector for each bond is computed and fed
    to the `StructuralCompletion` functions.
    """
    default_result = (np.array([0., 0., 0.0]), "origin")
    if molecule.GetNumberOfAtoms() == 0 or not atom_ids:
        return default_result

    # Unique but preserving order
    # (which set() does not do)
    unique_atom_ids = []
    for atom_id in atom_ids:
        if atom_id not in unique_atom_ids:
            unique_atom_ids.append(atom_id)

    number_of_bonds = len(unique_atom_ids) - 1
    if number_of_bonds not in completion_functions_to_try.keys():
        number_of_bonds = 0

    base_atom_id = unique_atom_ids[0]
    n_base_atom = molecule.GetAtom(base_atom_id).GetAtomicNumber()

    new_atom_radius = _get_covalent_radius(new_atom_n)
    base_atom_radius = _get_covalent_radius(n_base_atom)
    safety_distance = new_atom_radius + base_atom_radius

    validator = _create_molecule_validator(new_atom_n, molecule)

    known_positions = [_get_pos_array(atom_id, molecule) for atom_id in unique_atom_ids]

    def position_to_displacement(position: np.ndarray) -> np.ndarray:
        return position - known_positions[0]

    def displacement_to_position(displacement: np.ndarray) -> np.ndarray:
        return displacement + known_positions[0]

    def rescale_safe(displacement: np.ndarray) -> np.ndarray:
        length = np.sqrt(np.sum(displacement ** 2))
        if length == 0:
            return displacement  # We give up
        return displacement * safety_distance / length

    known_displacements = [position_to_displacement(p) for p in known_positions[1:]]

    for function, n_position_candidates in completion_functions_to_try[number_of_bonds]:
        displacement_candidates = _create_writeable_candidates(n_position_candidates)
        function(*known_displacements, *displacement_candidates)
        displacement_candidates_rescaled = [
            rescale_safe(d) for d in displacement_candidates
        ]

        position_candidates = [
            displacement_to_position(d) for d in displacement_candidates_rescaled
        ]
        for candidate in position_candidates:
            if validator(candidate.reshape((3,))):
                return candidate, function.__name__

    return default_result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MediatorPotential class.
"""
import numpy as np
from typing import Any, cast, Dict, List, Optional

from .custom_results import CustomResult


class MediatorPotential:
    """
    This class contains the mediator potential & the updates of the settings
    New results from the electronic_structure_calculation are stored here
    and used to approximate the gradient & energy
    All in atomic units
    """

    def __init__(
        self,
        molecule_version: int,
        calculation_result: CustomResult,
    ):
        self.__molecule_version = molecule_version
        self.__energy = calculation_result.energy
        self.__gradients = calculation_result.gradients
        self.__atomic_hessians = calculation_result.hessian
        self.__well_center = calculation_result.positions
        self.__charges = calculation_result.atomic_charges
        self.__bond_orders = calculation_result.bond_orders
        self.__molden_input = calculation_result.molden_input
        self.__settings = calculation_result.settings

    @property
    def molecule_version(self) -> int:
        return self.__molecule_version

    @property
    def molden_input(self) -> str:
        return self.__molden_input

    @property
    def settings(self) -> Dict[str, Any]:
        return self.__settings

    @property
    def bond_orders(self) -> Optional[np.ndarray]:
        return self.__bond_orders

    def get_gradients(self, current_positions: np.ndarray) -> np.ndarray:
        if self.__gradients is None:
            return np.zeros(shape=current_positions.shape)
        if self.__atomic_hessians is None:
            return self.__gradients
        # Returns approximate gradient
        distance = (
            np.array(current_positions).flatten()
            - np.array(self.__well_center).flatten()
        )
        approx_gradients = (
            np.array(self.__gradients).flatten()
            + np.dot(np.array(self.__atomic_hessians), distance)
        ).reshape((len(self.__gradients), 3))
        return approx_gradients

    def get_energy(self, current_positions: np.ndarray) -> float:
        if self.__atomic_hessians is None:
            return self.__energy
        # Returns the approx. energy according to the Taylor series
        # E = E0 + grad * (x-x0) + 0.5 * (x-x0)^T * hessian * (x-x0)
        distance = (
            np.array(current_positions).flatten()
            - np.array(self.__well_center).flatten()
        )
        approx_energy = (
            self.__energy
            + np.dot(np.array(self.__gradients).flatten(), distance)
            + 0.5
            * np.dot(
                np.dot(np.transpose(distance), np.array(self.__atomic_hessians)),
                distance,
            )
        )
        return cast(float, approx_energy)

    def get_atomic_charges(self) -> Optional[List[float]]:
        """Returns the atomic charges that were most recently computed by Sparrow."""

        return self.__charges



__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtCore import Qt
from PySide2.QtWidgets import (QCheckBox, QDialog, QDialogButtonBox,
                               QGridLayout, QHBoxLayout, QLabel, QLineEdit,
                               QSpinBox, QVBoxLayout)


class OpenMolcasOptionsPopup(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("OpenMolcas Interface Options")
        self.__layout = QGridLayout()
        self.__add_options()

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        # pylint: disable-next=E1101
        self.buttonBox.accepted.connect(self.accept)
        # pylint: disable-next=E1101
        self.buttonBox.rejected.connect(self.reject)
        self.__layout.addWidget(self.buttonBox, 1, 0, alignment=Qt.AlignCenter)

        self.setLayout(self.__layout)

    def __add_options(self):
        project_name_edit = QLineEdit(self)
        molcas_binary_location_edit = QLineEdit(self)
        molcas_flags_edit = QLineEdit(self)
        molcas_max_memory_edit = QSpinBox(self)
        molcas_scratch_dir_edit = QLineEdit(self)
        molcas_dumping_edit = QCheckBox(self)
        molcas_max_processes = QSpinBox(self)
        molcas_max_processes.setMinimum(1)

        layout = QHBoxLayout()
        sub_layout = QVBoxLayout()
        sub_layout.addWidget(QLabel("Project Name"))
        sub_layout.addWidget(QLabel("Binary Location"))
        sub_layout.addWidget(QLabel("<< pymolcas >> Flags"))
        sub_layout.addWidget(QLabel("Max Memory [MB]"))
        sub_layout.addWidget(QLabel("Scratch Location"))
        sub_layout.addWidget(QLabel("Dumping"))
        sub_layout.addWidget(QLabel("Max Processes"))
        layout.addLayout(sub_layout)
        sub_layout = QVBoxLayout()
        sub_layout.addWidget(project_name_edit)
        sub_layout.addWidget(molcas_binary_location_edit)
        sub_layout.addWidget(molcas_flags_edit)
        sub_layout.addWidget(molcas_max_memory_edit)
        sub_layout.addWidget(molcas_scratch_dir_edit)
        sub_layout.addWidget(molcas_dumping_edit)
        sub_layout.addWidget(molcas_max_processes)
        layout.addLayout(sub_layout)
        self.__layout.addLayout(layout, 0, 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""


from PySide2.QtWidgets import QProgressBar
from PySide2.QtGui import QPainter, QColor
from PySide2.QtCore import Qt

import scine_heron.config as config


class HeronProgressBar(QProgressBar):
    def paintEvent(self, event):
        super().paintEvent(event)
        if config.MODE in config.LIGHT_MODES:
            painter = QPainter(self)
            painter.setPen(QColor('#FFFFFF'))
            painter.drawText(self.rect(), Qt.AlignCenter, self.text())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the PlayStatusManager class.
"""

import typing
from PySide2.QtCore import QObject
if typing.TYPE_CHECKING:
    Signal = typing.Any
else:
    from PySide2.QtCore import Signal


class PlayStatusManager(QObject):
    """
    A class that manages an on/off status.
    """

    on_signal = Signal()
    off_signal = Signal()

    def __init__(self, parent: typing.Optional[QObject] = None):
        super().__init__(parent)
        self.__active = False

    def is_on(self) -> bool:
        """
        Returns True if and only if the status is currently "on"
        """
        return self.__active

    def __emit_if_necessary(self, status: bool) -> None:
        """
        Emits a signal if the status has changed from the provides status.
        """
        target = self.is_on()
        if status != target:
            if target:
                self.on_signal.emit()
            else:
                self.off_signal.emit()

    def stop(self) -> None:
        """
        Sets the status to "off".
        """
        status = self.is_on()
        self.__active = False
        self.__emit_if_necessary(status)

    def start(self) -> None:
        """
        Sets the status to "on".
        """
        status = self.is_on()
        self.__active = True
        self.__emit_if_necessary(status)
